// nvsts is a multi-directional shooter game.
// Copyright (C) 2008 Carles Creus Lopez
//
// nvsts is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>
// or write to the Free Software Foundation, Inc., 51 Franklin Street,
// Fifth Floor, Boston, MA 02110-1301 USA.
//
// You may contact the author at <crazycraft@gmail.com>

#include <cstdlib>
#include <SDL.h>
#include <boost/foreach.hpp>
#define foreach BOOST_FOREACH
#include "Config.hpp"
#include "Input.hpp"

#define JOY_THRESH 10922

// key identifier codes
enum keycodes {
	KESCAPE_CODE = 0,
	KIMM_CODE,
	KBOMB_CODE,
	KUP_CODE,
	KDOWN_CODE,
	KRIGHT_CODE,
	KLEFT_CODE,
	KFIREUP_CODE,
	KFIREDOWN_CODE,
	KFIRERIGHT_CODE,
	KFIRELEFT_CODE
};
// button identifier codes
enum buttoncodes {
	JESCAPE_CODE = 0,
	JIMM_CODE,
	JBOMB_CODE
};
// axis identifier codes
enum axiscodes {
	JUPDOWN_CODE = 0,
	JRIGHTLEFT_CODE,
	JFIREUPDOWN_CODE,
	JFIRERIGHTLEFT_CODE
};

// key mappings
std::map< SDLKey, std::pair<Input*,int> > Input::keymappings;
// joystick mappings
std::map< std::pair<int,int>, std::pair<Input*,int> > Input::buttonmappings;
std::map< std::pair<int,int>, std::pair<Input*,int> > Input::axismappings;

// opened joysticks
std::list<SDL_Joystick*> Input::joys;

// inputs for each player
std::vector<Input*> Input::inputs;
// aggregate input values
bool Input::agg_escape;
bool Input::agg_imm;
bool Input::agg_bomb;
util::Vector Input::agg_move;
util::Vector Input::agg_fire;

// initializes the input [depends on: Config]
void Input::init(){
	if ( SDL_NumJoysticks() ){
		// if joysticks presents, try to use them all
		SDL_JoystickEventState( SDL_ENABLE );

		int i = 0;
		do{
			SDL_Joystick *newj = SDL_JoystickOpen( i );
			if ( newj != NULL )
				joys.push_back( newj );
			++i;
		}while( i < SDL_NumJoysticks() );
	}

	reinit();
}
// reinitializes the input, call it when config changes
void Input::reinit(){
	// clearing all data
	keymappings.clear();
	buttonmappings.clear();
	axismappings.clear();

	// and rebuild it
	for( std::vector<Input*>::size_type i=0 ; i<Config::get_controls().size() ; ++i ){
		if ( inputs.size() == i )
			inputs.push_back( new Input() ); // create inputs as needed

		// adding all key mappings
		keymappings.insert( std::pair< SDLKey, std::pair<Input*,int> >(
			Config::get_controls()[i].kescape,
			std::pair<Input*,int>( inputs[i], KESCAPE_CODE ) ) );
		keymappings.insert( std::pair< SDLKey, std::pair<Input*,int> >(
			Config::get_controls()[i].kimm,
			std::pair<Input*,int>( inputs[i], KIMM_CODE ) ) );
		keymappings.insert( std::pair< SDLKey, std::pair<Input*,int> >(
			Config::get_controls()[i].kbomb,
			std::pair<Input*,int>( inputs[i], KBOMB_CODE ) ) );
		keymappings.insert( std::pair< SDLKey, std::pair<Input*,int> >(
			Config::get_controls()[i].kup,
			std::pair<Input*,int>( inputs[i], KUP_CODE ) ) );
		keymappings.insert( std::pair< SDLKey, std::pair<Input*,int> >(
			Config::get_controls()[i].kdown,
			std::pair<Input*,int>( inputs[i], KDOWN_CODE ) ) );
		keymappings.insert( std::pair< SDLKey, std::pair<Input*,int> >(
			Config::get_controls()[i].kright,
			std::pair<Input*,int>( inputs[i], KRIGHT_CODE ) ) );
		keymappings.insert( std::pair< SDLKey, std::pair<Input*,int> >(
			Config::get_controls()[i].kleft,
			std::pair<Input*,int>( inputs[i], KLEFT_CODE ) ) );
		keymappings.insert( std::pair< SDLKey, std::pair<Input*,int> >(
			Config::get_controls()[i].kfireup,
			std::pair<Input*,int>( inputs[i], KFIREUP_CODE ) ) );
		keymappings.insert( std::pair< SDLKey, std::pair<Input*,int> >(
			Config::get_controls()[i].kfiredown,
			std::pair<Input*,int>( inputs[i], KFIREDOWN_CODE ) ) );
		keymappings.insert( std::pair< SDLKey, std::pair<Input*,int> >(
			Config::get_controls()[i].kfireright,
			std::pair<Input*,int>( inputs[i], KFIRERIGHT_CODE ) ) );
		keymappings.insert( std::pair< SDLKey, std::pair<Input*,int> >(
			Config::get_controls()[i].kfireleft,
			std::pair<Input*,int>( inputs[i], KFIRELEFT_CODE ) ) );
		// adding all button mappings
		buttonmappings.insert( std::pair< std::pair<int,int>, std::pair<Input*,int> >(
			Config::get_controls()[i].jescape,
			std::pair<Input*,int>( inputs[i], JESCAPE_CODE ) ) );
		buttonmappings.insert( std::pair< std::pair<int,int>, std::pair<Input*,int> >(
			Config::get_controls()[i].jimm,
			std::pair<Input*,int>( inputs[i], JIMM_CODE ) ) );
		buttonmappings.insert( std::pair< std::pair<int,int>, std::pair<Input*,int> >(
			Config::get_controls()[i].jbomb,
			std::pair<Input*,int>( inputs[i], JBOMB_CODE ) ) );
		// adding all axis mappings
		axismappings.insert( std::pair< std::pair<int,int>, std::pair<Input*,int> >(
			Config::get_controls()[i].jupdown,
			std::pair<Input*,int>( inputs[i], JUPDOWN_CODE ) ) );
		axismappings.insert( std::pair< std::pair<int,int>, std::pair<Input*,int> >(
			Config::get_controls()[i].jrightleft,
			std::pair<Input*,int>( inputs[i], JRIGHTLEFT_CODE ) ) );
		axismappings.insert( std::pair< std::pair<int,int>, std::pair<Input*,int> >(
			Config::get_controls()[i].jfireupdown,
			std::pair<Input*,int>( inputs[i], JFIREUPDOWN_CODE ) ) );
		axismappings.insert( std::pair< std::pair<int,int>, std::pair<Input*,int> >(
			Config::get_controls()[i].jfirerightleft,
			std::pair<Input*,int>( inputs[i], JFIRERIGHTLEFT_CODE ) ) );

		// and saving the joystick inversions
		inputs[i]->jinvertupdown = Config::get_controls()[i].jinvertupdown;
		inputs[i]->jinvertrightleft = Config::get_controls()[i].jinvertrightleft;
		inputs[i]->jinvertfireupdown = Config::get_controls()[i].jinvertfireupdown;
		inputs[i]->jinvertfirerightleft = Config::get_controls()[i].jinvertfirerightleft;
	}

	aggregate();
}
// ends the input
void Input::finalize(){
	foreach( SDL_Joystick *j, joys )
		SDL_JoystickClose( j );
}
// starting a processing round
void Input::newround(){
	// input that is just a press (like bomb), is only remembered during one
	// single processing loop; else it would behave like really fast key
	// repetition

	agg_escape = false;
	agg_imm = false;
	agg_bomb = false;

	foreach( Input *i, inputs ){
		i->escape = false;
		i->imm = false;
		i->bomb = false;
	}
}
// compute the aggregates
void Input::aggregate(){
	// clearing data
	agg_escape = false;
	agg_imm = false;
	agg_bomb = false;
	agg_move = util::Vector(0.0, 0.0);
	agg_fire = util::Vector(0.0, 0.0);

	// adding each input data to the aggregates
	foreach( const Input *i, inputs ){
		agg_escape = agg_escape || i->escape;
		agg_imm = agg_imm || i->imm;
		agg_bomb = agg_bomb || i->bomb;
		agg_move += i->move;
		agg_fire += i->fire;
	}

	// normalizing vectors
	if ( agg_move != util::Vector(0.0, 0.0) )
		agg_move.magnitude(1.0);
	if ( agg_fire != util::Vector(0.0, 0.0) )
		agg_fire.magnitude(1.0);
}
// receiving key events
void Input::key( SDLKey key, bool pressed ){
	std::map< SDLKey, std::pair<Input*,int> >::const_iterator iter
		= keymappings.find( key );

	if ( iter == keymappings.end() )
		return;

	// sending event to the corresponding input
	((*iter).second).first->key( ((*iter).second).second, pressed );

	aggregate();
}
// receiving a key event
void Input::key( int code, bool pressed ){
	// updating state
	switch(code){
		case KESCAPE_CODE:
			escape = pressed;
			break;
		case KIMM_CODE:
			imm = pressed;
			break;
		case KBOMB_CODE:
			bomb = pressed;
			break;
		case KUP_CODE:
			up = pressed;
			break;
		case KDOWN_CODE:
			down = pressed;
			break;
		case KRIGHT_CODE:
			right = pressed;
			break;
		case KLEFT_CODE:
			left = pressed;
			break;
		case KFIREUP_CODE:
			fireup = pressed;
			break;
		case KFIREDOWN_CODE:
			firedown = pressed;
			break;
		case KFIRERIGHT_CODE:
			fireright = pressed;
			break;
		case KFIRELEFT_CODE:
			fireleft = pressed;
			break;
	}
	// updating corresponding vectors, if needed
	switch(code){
		case KUP_CODE:
		case KDOWN_CODE:
		case KRIGHT_CODE:
		case KLEFT_CODE:
			move = createVec( up, down, right, left );
			break;
		case KFIREUP_CODE:
		case KFIREDOWN_CODE:
		case KFIRERIGHT_CODE:
		case KFIRELEFT_CODE:
			fire = createVec( fireup, firedown, fireright, fireleft );
			break;
	}
}
// creates a vector representation
util::Vector Input::createVec( bool up, bool down, bool right, bool left ){
	real x = 0.0;
	real y = 0.0;

	if ( up )
		y += 1.0;
	if ( down )
		y -= 1.0;

	if ( right )
		x += 1.0;
	if ( left )
		x -= 1.0;

	// creating the normalized vector
	if ( (x == 0.0) && (y == 0.0) )
		return util::Vector(0.0, 0.0);
	else
		return util::Vector(x, y).magnitude(1.0);
}

// receiving button events
void Input::button( int which, int button, bool pressed ){
	std::map< std::pair<int,int>, std::pair<Input*,int> >::const_iterator iter
		= buttonmappings.find( std::pair<int,int>( which, button ) );

	if ( iter == buttonmappings.end() )
		return;

	// sending event to the corresponding input
	((*iter).second).first->button( ((*iter).second).second, pressed );

	aggregate();
}
// receiving a button event
void Input::button( int code, bool pressed ){
	// updating state
	switch(code){
		case JESCAPE_CODE:
			escape = pressed;
			break;
		case JIMM_CODE:
			imm = pressed;
			break;
		case JBOMB_CODE:
			bomb = pressed;
			break;
	}
}

// receiving axis events
void Input::axis( int which, int axis, int value ){
	std::map< std::pair<int,int>, std::pair<Input*,int> >::const_iterator iter
		= axismappings.find( std::pair<int,int>( which, axis ) );

	if ( iter == axismappings.end() )
		return;

	// sending event to the corresponding input
	((*iter).second).first->axis( ((*iter).second).second, value );

	aggregate();
}
// receiving an axis event
void Input::axis( int code, int value ){
	// updating state
	switch(code){
		case JUPDOWN_CODE:
			if ( jinvertupdown )
				updown = -value;
			else
				updown = value;
			break;
		case JRIGHTLEFT_CODE:
			if ( jinvertrightleft )
				rightleft = -value;
			else
				rightleft = value;
			break;
		case JFIREUPDOWN_CODE:
			if ( jinvertfireupdown )
				fireupdown = -value;
			else
				fireupdown = value;
			break;
		case JFIRERIGHTLEFT_CODE:
			if ( jinvertfirerightleft )
				firerightleft = -value;
			else
				firerightleft = value;
			break;
	}
	// updating corresponding vectors
	switch(code){
		case JUPDOWN_CODE:
		case JRIGHTLEFT_CODE:
			move = createVec( updown, rightleft );
			break;
		case JFIREUPDOWN_CODE:
		case JFIRERIGHTLEFT_CODE:
			fire = createVec( fireupdown, firerightleft );
			break;
	}
}
// creates a vector representation
util::Vector Input::createVec( int updown, int rightleft ){
	// if values do not surpass the threshold, ignore them
	if ( (std::abs(rightleft) < JOY_THRESH) && (std::abs(updown) < JOY_THRESH) )
		return util::Vector(0.0, 0.0);
	else
		return util::Vector(rightleft, updown).magnitude(1.0);
}

// ctor
Input::Input() :
 escape( false ),
 imm( false ),
 bomb( false ),
 move( 0.0, 0.0 ),
 fire( 0.0, 0.0 ),
 up( false ),
 down( false ),
 right( false ),
 left( false ),
 fireup( false ),
 firedown( false ),
 fireright( false ),
 fireleft( false ),
 updown( 0 ),
 rightleft( 0 ),
 fireupdown( 0 ),
 firerightleft( 0 ),
 jinvertupdown( false ),
 jinvertrightleft( false ),
 jinvertfireupdown( false ),
 jinvertfirerightleft( false )
{
}
// dtor
Input::~Input(){
}

// gets all input instances
const std::vector<Input*>& Input::get_inputs(){
	return inputs;
}
// gets an input instance
const Input *Input::get_input( int i ){
	return inputs[i];
}
// aggregates getters
bool Input::get_agg_escape(){
	return agg_escape;
}
bool Input::get_agg_imm(){
	return agg_imm;
}
bool Input::get_agg_bomb(){
	return agg_bomb;
}
const util::Vector& Input::get_agg_move(){
	return agg_move;
}
const util::Vector& Input::get_agg_fire(){
	return agg_fire;
}
// getters
bool Input::get_escape() const{
	return escape;
}
bool Input::get_imm() const{
	return imm;
}
bool Input::get_bomb() const{
	return bomb;
}
const util::Vector& Input::get_move() const{
	return move;
}
const util::Vector& Input::get_fire() const{
	return fire;
}
