// nvsts is a multi-directional shooter game.
// Copyright (C) 2008 Carles Creus Lopez
//
// nvsts is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>
// or write to the Free Software Foundation, Inc., 51 Franklin Street,
// Fifth Floor, Boston, MA 02110-1301 USA.
//
// You may contact the author at <crazycraft@gmail.com>

#ifndef MAIN_HPP
#define MAIN_HPP

#include <set>
#include <list>
#include <string>
#include <utility>
#include <SDL.h>
#include "util/Real.hpp"
#include "util/Vector.hpp"

class Main{
	public:
		// program start
		static void main( int argc, char *argv[] );
		// (re)starts the game
		static void start();
		// clears everything
		static void clear();
		// exit codes
		enum exitcodes {
			NORMAL_EXIT = 0,
			ERROR_INIT,
			ERROR_KREPEAT,
			ERROR_RESOLUTION,
			ERROR_DBUFFER,
			ERROR_AUDIO, // 5
			ERROR_AUDIO2,
			ERROR_FONT,
			ERROR_LOAD_RC,
			ERROR_PARSE_RC,
			ERROR_EXEC_RC, // 10
			ERROR_UNDEFINED_RC,
			ERROR_CONSTRAINTS_RC,
			ERROR_LOCALE
		};
		// quits the game
		static void quit( int code );

		// codes for different errors while unserializing
		typedef enum {
			UNSERIAL_SCORE = 1,
			UNSERIAL_LIFES,
			UNSERIAL_BOMBS,
			UNSERIAL_POWERUP_LIFES,
			UNSERIAL_REBIRTHWAIT, // 5
			UNSERIAL_REBIRTHWAIT_TIME,
			UNSERIAL_REBIRTHWAIT_ID,
			UNSERIAL_SPAWNTIME,
			UNSERIAL_SPAWNINGTIMES,
			UNSERIAL_SPAWNINGPOINTS, // 10
			UNSERIAL_SPAWNING_PROB,
			UNSERIAL_SHOT_POINTS,
			UNSERIAL_SHOT_COUNTER,
			UNSERIAL_PROCESSABLES,
			UNSERIAL_GRID_DIM, // 15
			UNSERIAL_ID,
			UNSERIAL_BLAST,
			UNSERIAL_RUBBLE,
			UNSERIAL_LAGVARD,
			UNSERIAL_DEBRIS, // 20
			UNSERIAL_SCARED,
			UNSERIAL_SPAWNSHIP,
			UNSERIAL_PLAYER_T,
			UNSERIAL_PLAYER_ID,
			UNSERIAL_PLAYER_LTIME, // 25
			UNSERIAL_PLAYER_FTIME,
			UNSERIAL_SHIELDED,
			UNSERIAL_REDHOLE_M,
			UNSERIAL_REDHOLE_FEED,
			UNSERIAL_REDHOLE_DAMAGE, // 30
			UNSERIAL_REDHOLE_MASSINCR,
			UNSERIAL_LASERPARTS_V_TAIL,
			UNSERIAL_LASERPARTS_V_GROW,
			UNSERIAL_LASERPARTS_PARTS,
			UNSERIAL_LASERPARTS_PARTS_LEFT, // 35
			UNSERIAL_LASER_LIFES,
			UNSERIAL_SPAWN_R,
			UNSERIAL_SPAWN_CHILDREN,
			UNSERIAL_EXAMINIS_REPRODUCE,
			UNSERIAL_EXAMINIS_WAITED, // 40
			UNSERIAL_EXAMINIS_APICULAS,
			UNSERIAL_EXAMINISAPICULA_LIFES,
			UNSERIAL_LIGHTNING_T,
			UNSERIAL_LIGHTNING_SPLIT,
			UNSERIAL_LIGHTNING_DEAD, // 45
			UNSERIAL_LIGHTNING_BRANCH,
			UNSERIAL_LIGHTNING_BODY,
			UNSERIAL_LIGHTNING_PRUNED,
			UNSERIAL_LIGHTNING_UNDEAD,
			UNSERIAL_LIGHTNINGSPARK_LIFES, // 50
			UNSERIAL_LIGHTNINGSPARK_TTL,
			UNSERIAL_LIGHTNINGSPARK_TURN,
			UNSERIAL_BULLET_LIFES
		} unserialize_error_t;
		// serializes the game
		static void serialize();
		// unserializes the game
		static void unserialize();

		// throws a bomb at xy
		static void bombing( const util::Vector& xy );

		// get game stats
		static int  get_score();
		static int  get_hiscore();
		static bool get_gameover();
		static int  get_lifes();
		static int  get_bombs();
		static bool get_merciless();
		// update game stats
		static void add_score( int scoreit );

		// computes game activity: 1.0=normal
		static real activity();

		// updates player information
		static void add_player( int id );
		static void remove_player( int id );

	private:
		// checks the arguments
		static void check_args( std::list<std::string> args );
		// program information
		static void print_notice();
		// starts game modules in proper order (see their dependencies)
		static void start_modules();
		// reading events since last round
		static void read_events();
		// draws everything (from back to front)
		static void draw();
		// process everything
		static void process( int milliseconds );
		// checks if the interval has already elapsed
		static bool timetrigger( int interval, int milliseconds, int& count );
		// player creation when presses bomb button
		static void entering_scheduler();
		// player creation when killed
		static void rebirth_scheduler( int milliseconds );

		// the window is not visible
		static bool hidden;
		// taking a screenshot?
		static bool screenshot;

		// enemy cleaning?
		static bool killenemies;

		// copy of rc::powerup_lifes
		static int powerup_lifes;

		// game stats
		static int  score;
		static int  hiscore;
		static bool gameover;
		static int  lifes;
		static int  bombs;
		static bool merciless;

		// processing intervals
		static int count_gravity;
		static int count_pursuit;
		static int count_dodging;
		static int count_collide;

		// active player's id
		static std::set<int> activeplayers; // unordered
		// scheduled players
		static std::list< std::pair<int,int> > rebirthwait;

		// computes the savegame file name
		static void fill_sg();
		// savegame file name
		static std::string sg;
};

#endif // Main.hpp
