// nvsts is a multi-directional shooter game.
// Copyright (C) 2008 Carles Creus Lopez
//
// nvsts is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>
// or write to the Free Software Foundation, Inc., 51 Franklin Street,
// Fifth Floor, Boston, MA 02110-1301 USA.
//
// You may contact the author at <crazycraft@gmail.com>

#ifndef AUTOMATON_HPP
#define AUTOMATON_HPP

#include <set>
#include <vector>
#include <utility>
#include <istream>
#include <ostream>
#include "../util/Partition.hpp"
#include "../util/Real.hpp"
#include "../util/Vector.hpp"

namespace au {

// prototyping
class Blast;
class Bullet;
class Debris;
class Examinis;
class ExaminisApicula;
class Lagvard;
class Laser;
class LaserParts;
class LaserTail;
class Lightning;
class LightningSpark;
class Player;
class Proton;
class RedHole;
class Scared;
class Shielded;
class Spawn;
class Spawned;
class Spawnship;
class Spring;

// base class of all automatons
class Automaton{
	public:
		// initializes the automaton [depends on: Constants]
		static void init();
		// ends the Automaton
		static void finalize();

		// serializes the automatons
		typedef enum {
			BLAST_SERIAL = 0,
			BULLET_SERIAL,
			DEBRIS_SERIAL,
			EXAMINIS_SERIAL,
			LAGVARD_SERIAL,
			LASER_SERIAL,
			LASERTAIL_SERIAL,
			LIGHTNING_SERIAL,
			PLAYER_SERIAL,
			PROTON_SERIAL,
			REDHOLE_SERIAL,
			SCARED_SERIAL,
			SHIELDED_SERIAL,
			SPAWN_SERIAL,
			SPAWNSHIP_SERIAL
		} serial_ids_t;
		static void sserialize( std::ostream& out );
		// unserializes the automatons
		static void sunserialize( std::istream& in );

		// drawing order for Visible objects (background first)
		enum drawingpriorities {
			NO_VISIBLE = 0,

			RUBBLE_PRIO,
			DEBRIS_PRIO,

			LASER_PRIO,
			LASERTAIL_PRIO,
			LIGHTNING_PRIO,
			BULLET_PRIO,

			LAGVARD_PRIO,
			SPAWNED_PRIO,
			SPAWNSHIP_PRIO,
			SHIELDED_PRIO,
			EXAMINIS_PRIO,
			SCARED_PRIO,
			PROTON_PRIO,
			REDHOLE_PRIO,

			PLAYER_PRIO
		};
		// drawing function
		static void draw();
#ifndef NDEBUG
		static void draw_map();
#endif
		// updates pursuiters
		static void pursuit();
		// process everything
		static void process( int milliseconds );
		// compute gravitational forces
		static void gravitate();
		// dodge shots
		static void dodge();
		// collider type options
		enum collideractive_t {
			NO_COLLIDER  = 0,

			BLAST_COL    = (1<<0),
			EXAMINIS_COL = (1<<1),
			PLAYER_COL   = (1<<2),
			REDHOLE_COL  = (1<<3),
			SHIELDED_COL = (1<<4),
			SHOT_COL     = (1<<5),

			COL_AMOUNT = 6
		};
		// compute collisions
		static void collide();

		// automaton killers
		static void kill_enemies();
		static void kill_players();
		static void kill_shots();
		static void kill_zombies();
		static void kill_remaining(); // callable when only blasts, debris, lasertails and lightnings remain

		// automaton amounts
		static int enemies_amount();
		static int players_amount();
		static int shots_amount();

		// checks if at xy with radius r there are no players
		static bool player_free_area( const util::Vector& xy, real r );

		// sets the shots to the starting value
		static void reset_shottype();
		// updates the shot type
		static void next_shottype( int score_incr );
		// gets the present milliseconds lapse between consecutive shots
		static int  shot_rate();
		// creates the present shot type at a radius distance from the origin in the given direction
		static void create_shot( const util::Vector& origin, const util::Vector& direction, real radius );
		// creates a tail for the laser
		static void create_lasertail( const util::Vector& xy, const util::Vector& v2, LaserParts *parts );
		// creates a lightning spark
		static LightningSpark *create_lightningspark( Lightning *swarmer, const util::Vector& xy,
		                                              const util::Vector& v, int lifes, int nttl, int turn );

		// enemies identifiers
		enum spawnablecodes {
			EXAMINIS_CODE = 0,
			LAGVARD_CODE,
			REDHOLE_CODE,
			SCARED_CODE,
			SHIELDED_CODE,
			SPAWNSHIP_CODE,

			SPAWNABLE_AMOUNT
		};
		// creates an enemy identified by code at location xy
		static void create_enemy( int code, const util::Vector& xy );
		// creates protons spawned from inside the redhole when it dies
		static void create_protons( const util::Vector& xy, real r, real m );
		// creates the spawnship's three spawned children
		static void create_spawned( const util::Vector& xy );
		// creates the examinis' apiculas
		static ExaminisApicula *create_examinisapicula( Examinis *swarmer, const util::Vector& xy, bool queen );

		// creates a player controled by user id
		static void create_player( int id );

		// creates a spring object at the given location
		static Spring *create_spring( real x, real y );

		// creates a blast at the given location
		static void create_blast( const util::Vector& xy );
		// litters the screen with some random-color debris, at a radius distance from xy in
		// omega direction (which has an imprecision of 2phi).
		// The amount of debris is max(1, ceil(3 2phi)), i.e. 3 per radiant, at least 1
		static void create_debris( const util::Vector& xy, real radius, real omega, real phi );
		// litters the screen with a circle of debris of the specified color, at a radius
		// distance from xy
		// The amount of debris is max(1, ceil(3 2phi)), i.e. 3 per radiant, at least 1
		static void create_debris( const util::Vector& xy, real radius, real red, real green, real blue );

	private:
		// visible automatons
		static std::set< std::pair<int,Automaton*> > visibles;
		// player following automatons
		static std::set<Automaton*> pursuiters; // unordered
		// while processing objects, processable objects may be created
		// so it needs to be deferred to avoid corrupting the iteration
		static bool deferproc;
		static std::set<Automaton*> deferredproc; // unordered
		// processable automatons
		static std::set<Automaton*> processables; // unordered
		// gravity listener automatons
		static std::set<Automaton*> gravitylisteners; // unordered
		// gravity provider automatons
		static std::set<Automaton*> gravityproviders; // unordered
		// shot dodgers
		static std::set<Automaton*> dodgers; // unordered
		// active collider automatons
		static std::set<Automaton*> collidersactive; // unordered
		// passive colliders automatons
		static util::Partition<Automaton*> *colliderspassive;
		// enemy automatons
		static std::set<Automaton*> enemies; // unordered
		// alive players
		static std::set<Automaton*> players; // unordered
		// shots
		static std::set<Automaton*> shots; // unordered
		// dead automatons
		static std::set<Automaton*> zombies; // unordered

		// options for shot_type
		enum shottypes {
			NORMAL_SHOTS = 0,
			FAST_SHOTS,
			MULTI_SHOTS,
			LASER_SHOTS,
			LIGHTNING_SHOTS,

			SHOTS_AMOUNT
		};
		// present type for shots
		static int shot_type;
		// counts the number of score points acumulated
		static int shot_counter;
		// score point intervals between each type of shot
		static int shot_points[];
		// modulus of the counter
		static int max_count;
		// shot rates for each type of shot
		static int shot_rates[];

		// litters the screen with some debris, at a radius distance from xy in omega direction
		// (which has an imprecision of 2phi).
		// The amount of debris is max(1, ceil(3 2phi)), i.e. 3 per radiant, at least 1
		static void create_debris( const util::Vector& xy, real radius, real omega, real phi,
		                           bool randcolors, real red, real green, real blue );

	// non-static part of Automaton

	public:
		// checks if this automaton intersects another automaton
		bool intersects_automaton( const Automaton *a ) const;
		// checks if this automaton intersects with the circle
		bool intersects_circle( const util::Vector& center, real radius ) const;
		// intersection magnitude of the two automatons
		real intersection_mag( const Automaton *a ) const;

		// return the vector of the distance from this automaton to a
		util::Vector distance_to( const Automaton *a ) const;

		// sets the final speeds of this automaton and automaton a when they ellastically
		// bounce and their masses are m1 and m2 respectively
		void ellastic_collision( Automaton *a, real m1, real m2, real nolosesratio );
		// moves the automatons to avoid the collision
		void avoid_collision( Automaton *a );
		// checks if a reflect of the automaton a is done and computes its resulting speed.
		// with pass_range is specified the angle that does not reflect, ~0.0 for perpendicular
		// and negative will always reflect
		bool reflect_collision( const au::Automaton *a, real pass_range, util::Vector& v2f ) const;
		// bouncing function
		virtual void bounce( const util::Vector& newv );

		// each collision type for active colliders
		virtual void recieve_collision( int& res, Bullet *p );
		virtual void recieve_collision( int& res, Debris *p );
		virtual void recieve_collision( int& res, ExaminisApicula *p );
		virtual void recieve_collision( int& res, Lagvard *p );
		virtual void recieve_collision( int& res, Laser *p );
		virtual void recieve_collision( int& res, LightningSpark *p );
		virtual void recieve_collision( int& res, Proton *p );
		virtual void recieve_collision( int& res, RedHole *p );
		virtual void recieve_collision( int& res, Scared *p );
		virtual void recieve_collision( int& res, Shielded *p );
		virtual void recieve_collision( int& res, Spawned *p );
		virtual void recieve_collision( int& res, Spawnship *p );

		// sets that it must not litter when dying
		void no_litter();

		// sets the position
		void set_xy( const util::Vector& newxy );
		// gets current position
		const util::Vector& get_xy() const;
		// gets mass
		real get_mass() const;

		// scoring points
		virtual int get_killscore() const;
		virtual int get_hitscore() const;

		// called when an automaton wants to commit suicided
		void add_zombie();

		// serializes the automaton
		virtual void serialize( std::ostream& out );

	protected:
		// ctor
		Automaton( const util::Vector& xy, const util::Vector& v, real r, int t = 0 );
		// dtor
		virtual ~Automaton();

		// helper function for processing code that sets the automaton inside the ring
		bool checkRing_stop();
		// helper function for processing code that sets the automaton inside the ring
		// but kills it if it was out
		bool checkRing_die();
		// helper function for processing code that kills the automaton if it was out
		// of the ring
		bool checkRing_justdie();
		// helper function for processing code that sets the automaton inside the ring
		// and makes it bounce if it has hit the ring limits
		bool checkRing_bounce( real nolosesratio );

		// position
		util::Vector xy;
		// velocity
		util::Vector v;
		// radius
		real r;
		// time elapsed since birth
		int t;

		// this object's mass
		real m;
		// gravity field
		util::Vector g;

		// whether or not it litters the screen when dies
		bool dolitter;

		// die when set_xy off the ring?
		bool ringdie;

		// starts the visible object
		void start_visible( int visiblityorder );
		// drawing function
		virtual void specific_draw();

		// starts the pursuiter object
		void start_pursuiter();
		// where is the nearest player. itis is an unit vector or zero.
		// distance is the distance or REAL_MAX
		virtual void specific_there( const util::Vector& itis, real distance );

		// starts the processable object
		void start_processable();
		// ends the processable object
		void end_processable();
		// process function
		virtual void specific_process( int milliseconds );

		// starts the gravity provider object
		void start_gravityprovider();
		// starts the gravity listener object
		void start_gravitylistener();

		// starts the dodger object
		void start_dodger();
		// dodger parameters
		virtual void specific_dodger_parameters( real& fr, real& fi ) const;
		// dodger function
		virtual void specific_dodger( const util::Vector& f );

		// starts the passive collider object
		// table with defined collisions:
		//
		//  Passive \ Active |  Blast  ExaminisApicula  Player  RedHole  Shielded  Shot
		//  -----------------|---------------------------------------------------------
		//           Bullet  |                x                    x        x
		//           Debris  |    x                                x
		//  ExaminisApicula  |                            x
		//          Lagvard  |                            x        x                x
		//            Laser  |                x                    x        x
		//   LightningSpark  |                x                    x        x
		//           Proton  |                            x                         x
		//          RedHole  |                            x        x
		//           Scared  |                            x        x                x
		//         Shielded  |                            x                 x
		//          Spawned  |                            x        x                x
		//        Spawnship  |                            x        x                x
		void start_passivecollider( int collided_types );
		// starts the active collider object
		void start_activecollider( int collideractive_type );
		// ends the active collider object
		void end_activecollider();
		// collision functions
		enum collision_t {
			NO_DEATHS    = 0x00,
			ACTIVE_DIES  = (1<<0),
			PASSIVE_DIES = (1<<1),
			BOTH_DIE     = PASSIVE_DIES | ACTIVE_DIES
		};
		virtual void send_collision( int& res, Automaton *a );

		// starts the enemy object
		void start_enemy( bool play );

		// starts the player object
		void start_player();

		// starts the shot object
		void start_shot();

	private:
		// helper function for processing code that checks if the automaton is
		// still in the ring and performs some actions if not:
		//   if move, sets it inside the ring changing xy
		//   if die, it kills the automaton
		//   if bounce, v direction is changed and multiplied by nolosesratio
		bool checkRing( bool move, bool die, bool bounce, real nolosesratio );

		// identification for automatons whose movement is tracked
		enum trackable_automatons {
			NO_TRACKABLE              = 0x00,
			TRACKABLE_COLLIDERPASSIVE = (1<<0)
		};
		// adds this type of trackability to the automaton, to properly qualify update calls
		void tracking_start( int type );
		// saves the information that needs to be tracked
		void tracking_save();
		// checks and notifies a change in the tracked data
		void tracking_check();
		// does this automaton need to be tracked? and its type
		int trackable;
		// tracking information
		util::Vector tracking_xy;
		real tracking_r;

		// this object's drawing priority
		int visiblityorder;

		// is it a pursuiter?
		bool pursuiter;

		// is it processable?
		bool processable;

		// is it a gravity provider?
		bool gravityprovider;
		// is it a gravity listener?
		bool gravitylistener;

		// is it a dodger?
		bool dodger;

		// type(s) of the objects this automaton collides to
		int collided_types;
		// type of the active collider
		int collideractive_type;

		// is it an enemy?
		bool enemy;

		// is it a player?
		bool player;

		// is it a shot?
		bool shot;
};

}

// getting prototype's definitions
#include "Blast.hpp"
#include "Bullet.hpp"
#include "Debris.hpp"
#include "Examinis.hpp"
#include "ExaminisApicula.hpp"
#include "Lagvard.hpp"
#include "Laser.hpp"
#include "LaserParts.hpp"
#include "LaserTail.hpp"
#include "Lightning.hpp"
#include "LightningSpark.hpp"
#include "Player.hpp"
#include "Proton.hpp"
#include "RedHole.hpp"
#include "Scared.hpp"
#include "Shielded.hpp"
#include "Spawn.hpp"
#include "Spawned.hpp"
#include "Spawnship.hpp"
#include "Spring.hpp"

#endif // Automaton.hpp
