// nvsts is a multi-directional shooter game.
// Copyright (C) 2008 Carles Creus Lopez
//
// nvsts is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>
// or write to the Free Software Foundation, Inc., 51 Franklin Street,
// Fifth Floor, Boston, MA 02110-1301 USA.
//
// You may contact the author at <crazycraft@gmail.com>

#include <cmath>
#include <boost/foreach.hpp>
#define foreach BOOST_FOREACH
#include "../rc/Constants.hpp"
#include "../util/Random.hpp"
#include "../vis/Draw.hpp"
#include "../Main.hpp"
#include "Examinis.hpp"

// ctor
au::Examinis::Examinis( const util::Vector& xy, int waited, int reproduce, bool create ) :
 Automaton( xy, util::Vector(0.0,0.0), rc::examinis_queen_r() ),
 queen( create ? Automaton::create_examinisapicula( this, xy, true ) : NULL ),
 apiculas(),
 waited( waited ),
 reproduce( reproduce )
{
	start_processable();
	start_pursuiter();
	start_visible( EXAMINIS_PRIO );
}
// dtor
au::Examinis::~Examinis(){
}
// remove the apicula whose dtor is being executed
void au::Examinis::takeout( ExaminisApicula *me ){
	// deleting
	if ( queen == me )
		queen = NULL;
	else
		apiculas.erase( me );

	// checking if nothing is left
	if ( (queen == NULL) && apiculas.empty() )
		delete this;
	else{
		// new reproduction priorities
		reproduce = compute_reproduce();
	}
}
// computes the reproduce time
int au::Examinis::compute_reproduce() const{
	// unbounded geometric progression
	return rc::examinis_t() * (1<<apiculas.size());
}
// serializes the automaton
void au::Examinis::serialize( std::ostream& out ){
	if ( queen == NULL )
		return;

	out << Automaton::EXAMINIS_SERIAL
	    << " " << xy
	    << " " << waited << " " << reproduce;

	out << " ";
	queen->serialize( out );

	out << " " << apiculas.size();
	foreach( ExaminisApicula *ea, apiculas ){
		out << " ";
		ea->serialize( out );
	}

	out << std::endl;
}
// unserializes the automaton
void au::Examinis::unserialize( std::istream& in ){
	util::Vector xy;
	int waited;
	int reproduce;

	in >> xy
	   >> waited >> reproduce;
	if ( reproduce < 0 )
		throw Main::UNSERIAL_EXAMINIS_REPRODUCE;
	if ( (waited < 0) || (waited >= reproduce) )
		throw Main::UNSERIAL_EXAMINIS_WAITED;

	Examinis *e = new Examinis( xy, waited, reproduce, false );
	e->queen = ExaminisApicula::unserialize( in, e, true );

	int as;
	in >> as;
	if ( as < 0 )
		throw Main::UNSERIAL_EXAMINIS_APICULAS;

	for( ; as ; --as ){
		e->apiculas.insert( ExaminisApicula::unserialize( in, e, false ) );
	}
}
// drawing function
void au::Examinis::specific_draw(){
	// drawing drones
	foreach( ExaminisApicula *ea, apiculas ){
		vis::Draw::reset();
		vis::Draw::set_color( rc::examinis_apicula_c_r(), rc::examinis_apicula_c_g(), rc::examinis_apicula_c_b() );
		vis::Draw::translate( ea->get_xy() );
		vis::Draw::rotate( ea->get_phase() );
		vis::Draw::play_record( vis::EXAMINIS_APICULA_ID );
	}

	// drawing queen
	if ( queen != NULL ){
		vis::Draw::reset();
		vis::Draw::set_color( rc::examinis_queen_c_r(), rc::examinis_queen_c_g(), rc::examinis_queen_c_b(),
		                      0.5 + static_cast<real>(queen->get_lifes())/(2*rc::examinis_queen_lifes()) );
		vis::Draw::translate( queen->get_xy() );
		vis::Draw::play_record( vis::EXAMINIS_QUEEN_ID );
	}
}
// where is the nearest player. itis is an unit vector or zero.
// distance is the distance or REAL_MAX
void au::Examinis::specific_there( const util::Vector& itis, real distance ){
	v = itis * rc::examinis_s();
}
// process function
void au::Examinis::specific_process( int milliseconds ){
	if ( queen == NULL ){
		// if queen has just been killed, kill all drones
		foreach( ExaminisApicula *ea, apiculas )
			ea->add_zombie();
	}else{
		// processing queen
		xy += v*milliseconds;
		checkRing_stop();
		queen->set_xy( xy );

		// spawning new drones
		waited += milliseconds;
		if ( waited >= reproduce ){
			apiculas.insert( Automaton::create_examinisapicula( this, xy, false ) );
			// new count
			waited = 0;
			reproduce = compute_reproduce();
		}

		// processing drones
		foreach( ExaminisApicula *ea, apiculas ){
			// getting coordinates in the ellipse
			real x = rc::examinis_orbitmin_r() * std::cos( ea->get_phase() ) * ea->get_orbit_ratiomax();
			real y = rc::examinis_orbitmin_r() * std::sin( ea->get_phase() );
			// creating vector
			util::Vector a_xy( x, y );
			// rotating vector
			a_xy.angle( a_xy.angle() + ea->get_phase2() );
			// setting position
			ea->set_xy( xy + a_xy );

			// updating phase
			ea->set_phase(  std::fmod(  ea->get_phase()+(rc::examinis_apicula_as()*milliseconds), static_cast<real>(2.0*PI) ) );
			// updating phase2
			ea->set_phase2( std::fmod( ea->get_phase2()+(rc::examinis_orbit_as()*milliseconds),   static_cast<real>(2.0*PI) ) );
		}
	}
}
