// nvsts is a multi-directional shooter game.
// Copyright (C) 2008 Carles Creus Lopez
//
// nvsts is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>
// or write to the Free Software Foundation, Inc., 51 Franklin Street,
// Fifth Floor, Boston, MA 02110-1301 USA.
//
// You may contact the author at <crazycraft@gmail.com>

#include <boost/foreach.hpp>
#define foreach BOOST_FOREACH
#include "../rc/Constants.hpp"
#include "../vis/Draw.hpp"
#include "../Main.hpp"
#include "LaserParts.hpp"

// ctor
au::LaserParts::LaserParts( real v_head, real v_tail ) :
 v_head( v_head ),
 v_tail( v_tail ),
 v_grow( v_head - v_tail ),
 l_full( 0.0 ),
 parts( 0 )
{
}
// dtor
au::LaserParts::~LaserParts(){
	parts.clear();
}
// processing function, updates the tail parameters as needed
void au::LaserParts::specific_process( int milliseconds, util::Vector& tail, util::Vector& vtail ){
	l_full += v_grow*milliseconds;

	if ( ! parts.empty() ){
		// if there are parts, the tail is approaching them so we need to decrease its left value
		foreach( laserpart_t& p, parts )
			p.left -= v_tail*milliseconds;

		// various parts may be gone, deleting them
		while( !parts.empty() && (parts.front().left <= 0.0) ){
			// changing parent laser position
			tail = parts.front().xy;
			// and setting its direction
			vtail.angle( parts.front().vtheta );

			parts.pop_front();
		}
	}
}
// drawing function
void au::LaserParts::specific_draw(){
	if ( l_full > 0.0 ){
		// draw all parts computing their alpha channel
		foreach( const laserpart_t& p, parts ){
			vis::Draw::set_color( rc::laser_c_r(), rc::laser_c_g(), rc::laser_c_b(), p.left/l_full );
			vis::Draw::add_point( p.xy );
		}
	}
}
// starts a part in the laser
void au::LaserParts::pushpart( const util::Vector& xy, real vtheta ){
	laserpart_t part;

	// we are at the head, so all the laser is left
	part.left   = l_full;
	// present position
	part.xy     = xy;
	// and direction the head is going after this stop point
	part.vtheta = vtheta;

	parts.push_back( part );
}
// there are no parts?
bool au::LaserParts::empty() const{
	return parts.empty();
}
// formatted io
std::ostream& operator<<( std::ostream& out, const au::LaserParts& lp ){
	out << lp.v_head << " " << lp.v_tail << " " << lp.l_full << " " << lp.parts.size();

	foreach( const au::LaserParts::laserpart_t& p, lp.parts ){
		out << " " << p.left
		    << " " << p.xy
		    << " " << p.vtheta;
	}

	return out;
}
std::istream& operator>>( std::istream& in, au::LaserParts& lp ){
	int lpps;

	in >> lp.v_head >> lp.v_tail >> lp.l_full >> lpps;
	lp.v_grow = lp.v_head - lp.v_tail;
	if ( lp.v_tail <= 0 )
		throw Main::UNSERIAL_LASERPARTS_V_TAIL;
	if ( lp.v_grow < 0 )
		throw Main::UNSERIAL_LASERPARTS_V_GROW;
	if ( lpps < 0 )
		throw Main::UNSERIAL_LASERPARTS_PARTS;

	real last = 0.0;
	lp.parts.clear();
	for( ; lpps ; --lpps ){
		au::LaserParts::laserpart_t p;

		in >> p.left
		   >> p.xy
		   >> p.vtheta;
		if ( (p.left <= 0) || (p.left < last) )
			throw Main::UNSERIAL_LASERPARTS_PARTS_LEFT;
		last = p.left;

		lp.parts.push_back( p );
	}

	return in;
}
