// nvsts is a multi-directional shooter game.
// Copyright (C) 2008 Carles Creus Lopez
//
// nvsts is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>
// or write to the Free Software Foundation, Inc., 51 Franklin Street,
// Fifth Floor, Boston, MA 02110-1301 USA.
//
// You may contact the author at <crazycraft@gmail.com>

#include <boost/foreach.hpp>
#define foreach BOOST_FOREACH
#include "../vis/Draw.hpp"
#include "../Audio.hpp"
#include "../Main.hpp"
#include "Lightning.hpp"

// some useful data types
typedef std::map< au::LightningSpark*, std::list<util::Vector> >::value_type bodyelem_t;

// ctor
au::Lightning::Lightning( const util::Vector& xy, const util::Vector& v, int t, int split, int dead, bool create, bool play ) :
 Automaton( xy, v, 0.0, t ),
 head( create ? Automaton::create_lightningspark( this, xy, v, rc::lightning_main_lifes(), t,
                static_cast<int>(util::Random::get(rc::lightning_turn_min(),rc::lightning_turn_max())) ) : NULL ),
 headbranch(),
 body(),
 pruned(),
 split( split ),
 dead( dead )
{
	start_processable();
	start_visible( LIGHTNING_PRIO );

	if ( create ){
		headbranch.push_back( head->get_xy() );
		headbranch.push_back( head->get_xy() );
	}

	if ( play )
		Audio::playSound( Audio::SOUND_LIGHTNING_FIRE );
}
// dtor
au::Lightning::~Lightning(){
}
// remove the spark whose dtor is being executed
void au::Lightning::takeout( LightningSpark *me ){
	if ( head == me ){
		// removing head
		head = NULL;
	}else{
		// removing body part
		pruned.push_back( body[me] );
		body.erase( me );
	}

	// checking if nothing is left
	if ( (head == NULL) && body.empty() )
		dead = rc::lightning_dead_t();
}
// processes a branch
void au::Lightning::process_branch( int milliseconds, LightningSpark *ls, std::list<util::Vector>& branch, bool ishead ){
	// decreasing its ttl
	int ttl = ls->get_ttl() - milliseconds;
	if ( ttl <= 0 )
		ls->add_zombie();
	else{
		ls->set_ttl( ttl );
		// checking a previous bounce
		if ( ls->has_bounced() ){
			branch.front() = ls->get_xy();
			branch.push_front( ls->get_xy() );
		}
		// movement direction
		int turn = ls->get_turn() - milliseconds;
		if ( turn <= 0 ){
			ls->set_turn( static_cast<int>(util::Random::get(rc::lightning_turn_min(),rc::lightning_turn_max())) );
			// changing its direction
			ls->drift( ls->get_v().angle() + util::Random::get(ishead ? rc::lightning_head_drift() : rc::lightning_drift()) );
			// updating its stop points
			branch.front() = ls->get_xy();
			branch.push_front( ls->get_xy() );
		}else
			ls->set_turn( turn );
		// updating its position
		ls->set_xy( ls->get_xy() + ls->get_v()*milliseconds );
	}
}
// draws a branch
void au::Lightning::draw_branch( const std::list<util::Vector>& branch, bool ishead ){
	// alpha channel
	real alpha = dead ? static_cast<real>(dead)/rc::lightning_dead_t() : 1.0;
	// line width and color
	if ( ishead ){
		vis::Draw::set_normal_lines();
		vis::Draw::set_color( rc::lightning_main_c_r(), rc::lightning_main_c_g(), rc::lightning_main_c_b(), alpha );
	}else{
		vis::Draw::set_thin_lines();
		vis::Draw::set_color( rc::lightning_c_r(), rc::lightning_c_g(), rc::lightning_c_b(), alpha );
	}

	// line painting
	vis::Draw::start_line_strip();
		int i = branch.size()-1;
		foreach( const util::Vector& v, branch ){
			if ( ishead && !i )
				vis::Draw::set_color( rc::lightning_main_c_r(), rc::lightning_main_c_g(), rc::lightning_main_c_b(), 0.0 );
			vis::Draw::add_point( v );
			--i;
		}
	vis::Draw::end_line_strip();
}
// serializes the automaton
void au::Lightning::serialize( std::ostream& out ){
	out << Automaton::LIGHTNING_SERIAL
	    << " " << xy << " " << v << " " << t
	    << " " << split << " " << dead;
	// head
	out << " " << (head != NULL);
	if ( head != NULL ){
		out << " ";
		head->serialize( out );
	}
	out << " " << headbranch.size();
	foreach( const util::Vector& v, headbranch )
		out << " " << v;
	// body
	out << " " << body.size();
	foreach( const bodyelem_t& b, body ){
		out << " ";
		b.first->serialize( out );
		out << " " << b.second.size();
		foreach( const util::Vector& v, b.second )
			out << " " << v;
	}
	// pruned
	out << " " << pruned.size();
	foreach( const std::list<util::Vector>& p, pruned ){
		out << " " << p.size();
		foreach( const util::Vector& v, p )
			out << " " << v;
	}
	out << std::endl;
}
// unserializes the automaton
void au::Lightning::unserialize( std::istream& in ){
	util::Vector xy;
	util::Vector v;
	int t;
	int split;
	int dead;

	in >> xy >> v >> t
	   >> split >> dead;
	if ( t < 0 )
		throw Main::UNSERIAL_LIGHTNING_T;
	if ( split <= 0 )
		throw Main::UNSERIAL_LIGHTNING_SPLIT;
	if ( (dead < 0) || (dead > rc::lightning_dead_t()) )
		throw Main::UNSERIAL_LIGHTNING_DEAD;

	Lightning *l = new Lightning( xy, v, t, split, dead, false, false );

	// head
	bool h;
	in >> h;
	if ( h ){
		l->head = LightningSpark::unserialize( in, l );
	}
	int hbs;
	in >> hbs;
	if ( hbs < 2 )
		throw Main::UNSERIAL_LIGHTNING_BRANCH;
	for( ; hbs ; --hbs ){
		util::Vector p;

		in >> p;

		l->headbranch.push_back( p );
	}
	// body
	int bs;
	in >> bs;
	if ( bs < 0 )
		throw Main::UNSERIAL_LIGHTNING_BODY;
	for( ; bs ; --bs ){
		LightningSpark *ls = LightningSpark::unserialize( in, l );

		int bbs;
		in >> bbs;
		if ( bbs < 2 )
			throw Main::UNSERIAL_LIGHTNING_BRANCH;

		std::list<util::Vector> bb;
		for( ; bbs ; --bbs ){
			util::Vector p;

			in >> p;

			bb.push_back( p );
		}

		l->body.insert( std::pair< LightningSpark*, std::list<util::Vector> >( ls, bb ) );
	}
	// pruned
	int ps;
	in >> ps;
	if ( ps < 0 )
		throw Main::UNSERIAL_LIGHTNING_PRUNED;
	for( ; ps ; --ps ){
		int pbs;
		in >> pbs;
		if ( pbs < 2 )
			throw Main::UNSERIAL_LIGHTNING_BRANCH;

		std::list<util::Vector> pb;
		for( ; pbs ; --pbs ){
			util::Vector p;

			in >> p;

			pb.push_back( p );
		}

		l->pruned.push_back( pb );
	}

	if ( dead && !((l->head == NULL) && l->body.empty()) )
		throw Main::UNSERIAL_LIGHTNING_UNDEAD;
}
// drawing function
void au::Lightning::specific_draw(){
	// drawing active branches
	foreach( bodyelem_t& b, body ){
		b.second.front() = b.first->get_xy();
		draw_branch( b.second, false );
	}
	// drawing dead branches
	foreach( const std::list<util::Vector>& p, pruned )
		draw_branch( p, false );
	// drawing main branch
	if ( head != NULL )
		headbranch.front() = head->get_xy();
	draw_branch( headbranch, true );
}
// process function
void au::Lightning::specific_process( int milliseconds ){
	if ( dead ){
		// if it is dead, we wait some time till self destruction
		dead -= milliseconds;
		if ( dead <= 0 )
			add_zombie();
	}else{
		// if it was still alive
		if ( head == NULL ){
			// but if it has been beheaded, then kill all parts
			foreach( const bodyelem_t& b, body )
				b.first->add_zombie();
		}else{
			// else it is still fully alive, so process the head
			process_branch( milliseconds, head, headbranch, true );

			// checking if we should do a split
			split -= milliseconds;
			if ( split <= 0 ){
				split = static_cast<int>(util::Random::get(rc::lightning_split_min(),rc::lightning_split_max()));
				// computing its velocity
				v.angle( head->get_v().angle() + util::Random::get(rc::lightning_split_angle()) );
				// computing its starting branch stop points
				std::list<util::Vector> branch;
				branch.push_back( head->get_xy() );
				branch.push_back( head->get_xy() );
				// creating
				body.insert( std::pair< LightningSpark*, std::list<util::Vector> >(
					Automaton::create_lightningspark( this, head->get_xy(), v, rc::lightning_lifes(), t,
					                                  static_cast<int>(util::Random::get(rc::lightning_turn_min(),rc::lightning_turn_max())) ),
					branch ) );
			}

			// and process each part
			foreach( bodyelem_t& b, body )
				process_branch( milliseconds, b.first, b.second, false );
		}
	}
}
