// nvsts is a multi-directional shooter game.
// Copyright (C) 2008 Carles Creus Lopez
//
// nvsts is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>
// or write to the Free Software Foundation, Inc., 51 Franklin Street,
// Fifth Floor, Boston, MA 02110-1301 USA.
//
// You may contact the author at <crazycraft@gmail.com>

#include <cmath>
#include <sstream>
#include <iomanip>
#include <algorithm>
#include "../au/Grid.hpp"
#include "../util/Text.hpp"
#include "../vis/Screen.hpp"
#include "../Config.hpp"
#include "../Locale.hpp"
#include "Menu.hpp"
#include "Menu_Config.hpp"

#define VALUES_FIELD 9

// line codes
enum config_lines {
	CONFIG_LINE_LANG = 0,
	CONFIG_LINE_SKIP0,
	CONFIG_LINE_GRID,
	CONFIG_LINE_GRIDBEZIER,
	CONFIG_LINE_SKIP1,
	CONFIG_LINE_DEBRIS,
	CONFIG_LINE_SKIP2,
	CONFIG_LINE_AUDIO,
	CONFIG_LINE_SKIP3,
	CONFIG_LINE_RESOLUTION,
	CONFIG_LINE_BITDEPTH,
	CONFIG_LINE_FULLSCREEN,
	CONFIG_LINE_LINESMOOTH,
	CONFIG_LINE_SKIP4,
	CONFIG_LINE_OK,

	CONFIG_LINE_MIN = CONFIG_LINE_LANG,
	CONFIG_LINE_MAX = CONFIG_LINE_OK
};

// ctor
gui::Menu_Config::Menu_Config(){
	text = std::vector<std::string>(CONFIG_LINE_MAX+1);
	text[CONFIG_LINE_SKIP0] = " ";
	text[CONFIG_LINE_SKIP1] = " ";
	text[CONFIG_LINE_SKIP2] = " ";
	text[CONFIG_LINE_SKIP3] = " ";
	text[CONFIG_LINE_SKIP4] = " ";
}
// dtor
gui::Menu_Config::~Menu_Config(){
}
// config pane start
void gui::Menu_Config::start( int from ){
	line = CONFIG_LINE_OK;

	// getting present configuration

	resx       = Config::get_resx();
	resy       = Config::get_resy();
	bitdepth   = Config::get_bitdepth();
	fullscreen = Config::get_fullscreen();
	linesmooth = Config::get_linesmooth();

	grid       = Config::get_grid();
	gridrows   = Config::get_gridrows();
	gridcols   = Config::get_gridcols();
	gridbezier = Config::get_gridbezier();

	debris = Config::get_debris();

	audio = Config::get_audio();

	update();
}
// config pane ended
void gui::Menu_Config::end(){
	Menu::start( Menu::ROOT_PANE );
}
// updates the text
void gui::Menu_Config::update(){
	// grid text
	std::ostringstream line_grid_aux;
	if ( grid )
		line_grid_aux << gridcols << 'x' << gridrows;
	else
		line_grid_aux << Locale::get( Locale::STR_OFF ) << fixwidth( Locale::get( Locale::STR_OFF ) );
	std::ostringstream line_grid;
	line_grid
		<< Locale::get( Locale::CONFIG_STR_GRID )
		<< std::setfill(' ') << std::setw(VALUES_FIELD) << std::left
		<< line_grid_aux.str();

	// grid bezier text
	std::ostringstream line_gridbezier;
	line_gridbezier
		<< Locale::get( Locale::CONFIG_STR_GRIDBEZIER )
		<< std::setfill(' ') << std::setw(VALUES_FIELD) << std::left
		<< Locale::get( (grid && gridbezier) ? Locale::STR_ON : Locale::STR_OFF )
		<< fixwidth( Locale::get( (grid && gridbezier) ? Locale::STR_ON : Locale::STR_OFF ) );

	// debris text
	std::ostringstream line_debris;
	line_debris
		<< Locale::get( Locale::CONFIG_STR_DEBRIS )
		<< std::setfill(' ') << std::setw(VALUES_FIELD) << std::left
		<< Locale::get( debris ? Locale::STR_ON : Locale::STR_OFF ) << fixwidth( Locale::get( debris ? Locale::STR_ON : Locale::STR_OFF ) );

	// audio text
	std::ostringstream line_audio;
	line_audio
		<< Locale::get( Locale::CONFIG_STR_AUDIO )
		<< std::setfill(' ') << std::setw(VALUES_FIELD) << std::left;
		if ( audio )
			line_audio << audio;
		else
			line_audio << Locale::get( Locale::STR_OFF ) << fixwidth( Locale::get( Locale::STR_OFF ) );

	// resolution text
	std::ostringstream line_resolution_aux;
	line_resolution_aux << resx << 'x' << resy;
	std::ostringstream line_resolution;
	line_resolution
		<< Locale::get( Locale::CONFIG_STR_RESOLUTION )
		<< std::setfill(' ') << std::setw(VALUES_FIELD) << std::left
		<< line_resolution_aux.str();

	// bitdepth text
	std::ostringstream line_bitdepth;
	line_bitdepth
		<< Locale::get( Locale::CONFIG_STR_BITDEPTH )
		<< std::setfill(' ') << std::setw(VALUES_FIELD) << std::left
		<< bitdepth;

	// fullscreen text
	std::ostringstream line_fullscreen;
	line_fullscreen
		<< Locale::get( Locale::CONFIG_STR_FULLSCREEN )
		<< std::setfill(' ') << std::setw(VALUES_FIELD) << std::left
		<< Locale::get( fullscreen ? Locale::STR_ON : Locale::STR_OFF ) << fixwidth( Locale::get( fullscreen ? Locale::STR_ON : Locale::STR_OFF ) );

	// antialias text
	std::ostringstream line_linesmooth;
	line_linesmooth
		<< Locale::get( Locale::CONFIG_STR_LINESMOOTH )
		<< std::setfill(' ') << std::setw(VALUES_FIELD) << std::left
		<< Locale::get( linesmooth ? Locale::STR_ON : Locale::STR_OFF ) << fixwidth( Locale::get( linesmooth ? Locale::STR_ON : Locale::STR_OFF ) );

	// saving the results
	text[CONFIG_LINE_LANG]       = Config::get_lang();
	text[CONFIG_LINE_GRID]       = line_grid.str();
	text[CONFIG_LINE_GRIDBEZIER] = line_gridbezier.str();
	text[CONFIG_LINE_DEBRIS]     = line_debris.str();
	text[CONFIG_LINE_AUDIO]      = line_audio.str();
	text[CONFIG_LINE_RESOLUTION] = line_resolution.str();
	text[CONFIG_LINE_BITDEPTH]   = line_bitdepth.str();
	text[CONFIG_LINE_FULLSCREEN] = line_fullscreen.str();
	text[CONFIG_LINE_LINESMOOTH] = line_linesmooth.str();
	text[CONFIG_LINE_OK]         = Locale::get( Locale::STR_OK );

	Menu::update();
}
// save config
void gui::Menu_Config::save(){
	// saving screen options only if they have changed
	if ( (resx != Config::get_resx())
	     || (resy != Config::get_resy())
	     || (bitdepth != Config::get_bitdepth())
	     || (fullscreen != Config::get_fullscreen())
	     || (linesmooth != Config::get_linesmooth())
	   ){
		Config::set_resx( resx );
		Config::set_resy( resy );
		Config::set_bitdepth( bitdepth );
		Config::set_fullscreen( fullscreen );
		Config::set_linesmooth( linesmooth );
		vis::Screen::init();
	}

	// saving the rest of configuration

	Config::set_grid( grid );
	Config::set_gridrows( gridrows );
	Config::set_gridcols( gridcols );
	Config::set_gridbezier( gridbezier );
	au::Grid::reinit();

	Config::set_debris( debris );

	Config::set_audio( audio );
}
// user event: exit
void gui::Menu_Config::exit(){
	end();
}
// user event: movement
void gui::Menu_Config::move( int updown ){
	// changing line
	line += updown;
	switch(line){
		case CONFIG_LINE_SKIP0:
		case CONFIG_LINE_SKIP1:
		case CONFIG_LINE_SKIP2:
		case CONFIG_LINE_SKIP3:
		case CONFIG_LINE_SKIP4:
			line += updown;
			break;
	}
	// wrapping
	if ( line < CONFIG_LINE_MIN )
		line = CONFIG_LINE_MAX;
	else if ( line > CONFIG_LINE_MAX )
		line = CONFIG_LINE_MIN;
}
// user event: change
void gui::Menu_Config::change( int moreless ){
	std::set<std::string>::const_iterator iter1;
	std::set< std::pair<int,int> >::const_iterator iter2;

	switch(line){
		// language change
		case CONFIG_LINE_LANG:
			// getting next one, language list wraps at its ends
			iter1 = Locale::get_availablelangs().find( Config::get_lang() );
			if ( (moreless > 0) || (iter1 != Locale::get_availablelangs().begin()) ){
				if ( moreless > 0 )
					iter1++;
				else
					iter1--;
				if ( iter1 != Locale::get_availablelangs().end() )
					Config::set_lang( *iter1 );
				else
					Config::set_lang( *Locale::get_availablelangs().begin() );
			}else
				Config::set_lang( *Locale::get_availablelangs().rbegin() );
			Locale::reinit();
			break;
		// grid change
		case CONFIG_LINE_GRID:
			if ( (moreless < 0) && (gridrows == ROWS_MIN) )
				grid = false; // deactivating the grid
			else{
				if ( grid ){ // modifying only if grid was activated
					gridrows = std::min( ROWS_MAX, gridrows+moreless );
					gridcols = static_cast<int>( std::ceil( gridrows*COLS_MUL ) );
				}
				grid = true;
			}
			break;
		// bezier change
		case CONFIG_LINE_GRIDBEZIER:
			if ( grid )
				gridbezier = ! gridbezier;
			break;
		// debris change
		case CONFIG_LINE_DEBRIS:
			debris = ! debris;
			break;
		// audio change
		case CONFIG_LINE_AUDIO:
			audio = std::max( 0, std::min( MAX_VOLUME, audio+moreless ) );
			break;
		// resolution change
		case CONFIG_LINE_RESOLUTION:
			// getting next one
			iter2 = vis::Screen::get_availableres().find( std::pair<int,int>(resx, resy) );
			if ( (moreless > 0) || (iter2 != vis::Screen::get_availableres().begin()) ){
				if ( moreless > 0 )
					iter2++;
				else
					iter2--;
				if ( iter2 != vis::Screen::get_availableres().end() ){
					resx = (*iter2).first;
					resy = (*iter2).second;
				}
			}
			break;
		// bitdepth change
		case CONFIG_LINE_BITDEPTH:
			bitdepth = std::min( 32, std::max( 8, ((bitdepth/8)+moreless)*8 ) );
			break;
		// fullscreen change
		case CONFIG_LINE_FULLSCREEN:
			fullscreen = ! fullscreen;
			break;
		// antialias change
		case CONFIG_LINE_LINESMOOTH:
			linesmooth = ! linesmooth;
			break;
	}
	// update the text if configuration has been changed
	switch(line){
		case CONFIG_LINE_LANG:
		case CONFIG_LINE_GRID:
		case CONFIG_LINE_GRIDBEZIER:
		case CONFIG_LINE_DEBRIS:
		case CONFIG_LINE_AUDIO:
		case CONFIG_LINE_RESOLUTION:
		case CONFIG_LINE_BITDEPTH:
		case CONFIG_LINE_FULLSCREEN:
		case CONFIG_LINE_LINESMOOTH:
			update();
			break;
	}
}
// user event: accept
void gui::Menu_Config::accept(){
	switch(line){
		// ending
		case CONFIG_LINE_OK:
			save();
			end();
			break;
	}
}
// capture ended: ignore
void gui::Menu_Config::notify( bool success, SDLKey key, int joy, int axis_but, bool invert ){
}
// returns the space to append to correct str's width
std::string gui::Menu_Config::fixwidth( const std::string& str ){
	return std::string( util::Text::utf8OverheadBytes(str), ' ' );
}
