// nvsts is a multi-directional shooter game.
// Copyright (C) 2008 Carles Creus Lopez
//
// nvsts is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>
// or write to the Free Software Foundation, Inc., 51 Franklin Street,
// Fifth Floor, Boston, MA 02110-1301 USA.
//
// You may contact the author at <crazycraft@gmail.com>

#include <sstream>
#include <iomanip>
#include <algorithm>
#include <boost/foreach.hpp>
#define foreach BOOST_FOREACH
#include "../Input.hpp"
#include "../Config.hpp"
#include "../Locale.hpp"
#include "Menu.hpp"
#include "Menu_Controls.hpp"

#define NAMES_FIELD 11

// line codes
enum controls_lines {
	CONTROLS_LINE_PLAYER = 0,
	CONTROLS_LINE_SKIP0,
	CONTROLS_LINE_KEYBOARD,
	CONTROLS_LINE_KESCAPE,
	CONTROLS_LINE_KBOMB,
	CONTROLS_LINE_KUP,
	CONTROLS_LINE_KDOWN,
	CONTROLS_LINE_KRIGHT,
	CONTROLS_LINE_KLEFT,
	CONTROLS_LINE_KFIREUP,
	CONTROLS_LINE_KFIREDOWN,
	CONTROLS_LINE_KFIRERIGHT,
	CONTROLS_LINE_KFIRELEFT,
	CONTROLS_LINE_SKIP1,
	CONTROLS_LINE_JOYSTICK,
	CONTROLS_LINE_JESCAPE,
	CONTROLS_LINE_JBOMB,
	CONTROLS_LINE_JUPDOWN,
	CONTROLS_LINE_JRIGHTLEFT,
	CONTROLS_LINE_JFIREUPDOWN,
	CONTROLS_LINE_JFIRERIGHTLEFT,
	CONTROLS_LINE_SKIP2,
	CONTROLS_LINE_OK,

	// unique integer identifiers for not visible controls
	SPECIAL_LINE_ID_KIMM,
	SPECIAL_LINE_ID_JIMM,

	CONTROLS_LINE_MIN = CONTROLS_LINE_PLAYER,
	CONTROLS_LINE_MAX = CONTROLS_LINE_OK
};

// ctor
gui::Menu_Controls::Menu_Controls(){
	text = std::vector<std::string>(CONTROLS_LINE_MAX+1);
	text[CONTROLS_LINE_SKIP0] = " ";
	text[CONTROLS_LINE_SKIP1] = " ";
	text[CONTROLS_LINE_SKIP2] = " ";
}
// dtor
gui::Menu_Controls::~Menu_Controls(){
}
// controls pane start
void gui::Menu_Controls::start( int from ){
	line = CONTROLS_LINE_OK;

	playerid = 0;
	playercntrls = Config::get_controls();
	// building mappings, from last player to first one
	for( int i=static_cast<int>(playercntrls.size())-1 ; i>=0 ; --i ){
		// setting up key mappings
		addkeymapping( playercntrls[i].kescape,    std::pair<int,int>( i, CONTROLS_LINE_KESCAPE ) );
		addkeymapping( playercntrls[i].kimm,       std::pair<int,int>( i, SPECIAL_LINE_ID_KIMM ) );
		addkeymapping( playercntrls[i].kbomb,      std::pair<int,int>( i, CONTROLS_LINE_KBOMB ) );
		addkeymapping( playercntrls[i].kup,        std::pair<int,int>( i, CONTROLS_LINE_KUP ) );
		addkeymapping( playercntrls[i].kdown,      std::pair<int,int>( i, CONTROLS_LINE_KDOWN ) );
		addkeymapping( playercntrls[i].kright,     std::pair<int,int>( i, CONTROLS_LINE_KRIGHT ) );
		addkeymapping( playercntrls[i].kleft,      std::pair<int,int>( i, CONTROLS_LINE_KLEFT ) );
		addkeymapping( playercntrls[i].kfireup,    std::pair<int,int>( i, CONTROLS_LINE_KFIREUP ) );
		addkeymapping( playercntrls[i].kfiredown,  std::pair<int,int>( i, CONTROLS_LINE_KFIREDOWN ) );
		addkeymapping( playercntrls[i].kfireright, std::pair<int,int>( i, CONTROLS_LINE_KFIRERIGHT ) );
		addkeymapping( playercntrls[i].kfireleft,  std::pair<int,int>( i, CONTROLS_LINE_KFIRELEFT ) );
		// setting up joystick button mappings
		addbuttonmapping( playercntrls[i].jescape, std::pair<int,int>( i, CONTROLS_LINE_JESCAPE ) );
		addbuttonmapping( playercntrls[i].jimm,    std::pair<int,int>( i, SPECIAL_LINE_ID_JIMM ) );
		addbuttonmapping( playercntrls[i].jbomb,   std::pair<int,int>( i, CONTROLS_LINE_JBOMB ) );
		// setting up joystick axis mappings
		addaxismapping( playercntrls[i].jupdown,        std::pair<int,int>( i, CONTROLS_LINE_JUPDOWN ) );
		addaxismapping( playercntrls[i].jrightleft,     std::pair<int,int>( i, CONTROLS_LINE_JRIGHTLEFT ) );
		addaxismapping( playercntrls[i].jfireupdown,    std::pair<int,int>( i, CONTROLS_LINE_JFIREUPDOWN ) );
		addaxismapping( playercntrls[i].jfirerightleft, std::pair<int,int>( i, CONTROLS_LINE_JFIRERIGHTLEFT ) );
	}

	resetplcntrl();
	update();
}
// sets up a key and fixes configuration to avoid key remappings
void gui::Menu_Controls::addkeymapping( SDLKey key, const std::pair<int,int>& id ){
	// nothing to do if present key is not useful information
	if ( key == SDLK_UNKNOWN )
		return;

	// if the mapping already existed
	std::map< SDLKey, std::pair<int,int> >::iterator iter = keymappings.find( key );
	if ( iter != keymappings.end() ){
		// nothing to do if it is exactly for same player and same key
		if ( (*iter).second == id )
			return;

		// else we need to change present configuration
		switch((*iter).second.second){
			case CONTROLS_LINE_KESCAPE:
				playercntrls[(*iter).second.first].kescape = SDLK_UNKNOWN;
				break;
			case CONTROLS_LINE_KBOMB:
				playercntrls[(*iter).second.first].kbomb = SDLK_UNKNOWN;
				break;
			case CONTROLS_LINE_KUP:
				playercntrls[(*iter).second.first].kup = SDLK_UNKNOWN;
				break;
			case CONTROLS_LINE_KDOWN:
				playercntrls[(*iter).second.first].kdown = SDLK_UNKNOWN;
				break;
			case CONTROLS_LINE_KRIGHT:
				playercntrls[(*iter).second.first].kright = SDLK_UNKNOWN;
				break;
			case CONTROLS_LINE_KLEFT:
				playercntrls[(*iter).second.first].kleft = SDLK_UNKNOWN;
				break;
			case CONTROLS_LINE_KFIREUP:
				playercntrls[(*iter).second.first].kfireup = SDLK_UNKNOWN;
				break;
			case CONTROLS_LINE_KFIREDOWN:
				playercntrls[(*iter).second.first].kfiredown = SDLK_UNKNOWN;
				break;
			case CONTROLS_LINE_KFIRERIGHT:
				playercntrls[(*iter).second.first].kfireright = SDLK_UNKNOWN;
				break;
			case CONTROLS_LINE_KFIRELEFT:
				playercntrls[(*iter).second.first].kfireleft = SDLK_UNKNOWN;
				break;
			case SPECIAL_LINE_ID_KIMM:
				playercntrls[(*iter).second.first].kimm = SDLK_UNKNOWN;
				break;
		}
		// and erase present mapping
		keymappings.erase( iter );
	}

	// adding the mapping
	keymappings.insert( std::pair< SDLKey, std::pair<int,int> >( key, id ) );
}
// sets up a button and fixes configuration to avoid button remappings
void gui::Menu_Controls::addbuttonmapping( const std::pair<int,int>& joybut, const std::pair<int,int>& id ){
	// nothing to do if present button is not useful information
	if ( joybut == std::pair<int,int>(-1, -1) )
		return;

	// if the mapping already existed
	std::map< std::pair<int,int>, std::pair<int,int> >::iterator iter = buttonmappings.find( joybut );
	if ( iter != buttonmappings.end() ){
		// nothing to do if it is exactly for same player and same button
		if ( (*iter).second == id )
			return;

		// else we need to change present configuration
		switch((*iter).second.second){
			case CONTROLS_LINE_JESCAPE:
				playercntrls[(*iter).second.first].jescape = std::pair<int,int>(-1, -1);
				break;
			case CONTROLS_LINE_JBOMB:
				playercntrls[(*iter).second.first].jbomb = std::pair<int,int>(-1, -1);
				break;
			case SPECIAL_LINE_ID_JIMM:
				playercntrls[(*iter).second.first].jimm = std::pair<int,int>(-1, -1);
				break;
		}
		// and erase present mapping
		buttonmappings.erase( iter );
	}

	// adding the mapping
	buttonmappings.insert( std::pair< std::pair<int,int>, std::pair<int,int> >( joybut, id ) );
}
// sets up an axis and fixes configuration to avoid axis remappings
void gui::Menu_Controls::addaxismapping( const std::pair<int,int>& joyaxis, const std::pair<int,int>& id ){
	// nothing to do if present axis is not useful information
	if ( joyaxis == std::pair<int,int>(-1, -1) )
		return;

	// if the mapping already existed
	std::map< std::pair<int,int>, std::pair<int,int> >::iterator iter = axismappings.find( joyaxis );
	if ( iter != axismappings.end() ){
		// nothing to do if it is exactly for same player and same axis
		if ( (*iter).second == id )
			return;

		// else we need to change present configuration
		switch((*iter).second.second){
			case CONTROLS_LINE_JUPDOWN:
				playercntrls[(*iter).second.first].jupdown = std::pair<int,int>(-1, -1);
				break;
			case CONTROLS_LINE_JRIGHTLEFT:
				playercntrls[(*iter).second.first].jrightleft = std::pair<int,int>(-1, -1);
				break;
			case CONTROLS_LINE_JFIREUPDOWN:
				playercntrls[(*iter).second.first].jfireupdown = std::pair<int,int>(-1, -1);
				break;
			case CONTROLS_LINE_JFIRERIGHTLEFT:
				playercntrls[(*iter).second.first].jfirerightleft = std::pair<int,int>(-1, -1);
				break;
		}
		// and erase present mapping
		axismappings.erase( iter );
	}

	// adding the mapping
	axismappings.insert( std::pair< std::pair<int,int>, std::pair<int,int> >( joyaxis, id ) );
}
// controls pane ended
void gui::Menu_Controls::end(){
	// clearing data for next start
	keymappings.clear();
	buttonmappings.clear();
	axismappings.clear();
	// returning to root pane
	Menu::start( Menu::ROOT_PANE );
}
// gets a string representation of a key
std::string gui::Menu_Controls::getname( SDLKey key ){
	if ( key == SDLK_UNKNOWN )
		return std::string("");
	else
		return std::string( SDL_GetKeyName(key) );
}
// gets a string representation of a joystic-axis or a joystick-button
std::string gui::Menu_Controls::getname( bool axis, const std::pair<int,int>& joy ){
	if ( (joy.first<0) || (joy.second<0) )
		return std::string("");
	else{
		std::ostringstream res;
		res << "joy" << joy.first << "-";
		if ( axis )
			res << "axis" << joy.second;
		else
			res << "but" << joy.second;
		return res.str();
	}
}
// updates the text
void gui::Menu_Controls::update(){
	// player text
	std::ostringstream line_player;
	line_player << Locale::get( Locale::CONTROLS_STR_PLAYER ) << " " << (playerid+1);

	// keyboard escape
	std::ostringstream line_kescape;
	line_kescape
		<< Locale::get( Locale::CONTROLS_STR_KESCAPE )
		<< std::setfill(' ') << std::setw(NAMES_FIELD) << std::left
		<< getname( plcntrl.kescape );
	// keyboard bomb
	std::ostringstream line_kbomb;
	line_kbomb
		<< Locale::get( Locale::CONTROLS_STR_KBOMB )
		<< std::setfill(' ') << std::setw(NAMES_FIELD) << std::left
		<< getname( plcntrl.kbomb );
	// keyboard up
	std::ostringstream line_kup;
	line_kup
		<< Locale::get( Locale::CONTROLS_STR_KUP )
		<< std::setfill(' ') << std::setw(NAMES_FIELD) << std::left
		<< getname( plcntrl.kup );
	// keyboard down
	std::ostringstream line_kdown;
	line_kdown
		<< Locale::get( Locale::CONTROLS_STR_KDOWN )
		<< std::setfill(' ') << std::setw(NAMES_FIELD) << std::left
		<< getname( plcntrl.kdown );
	// keyboard right
	std::ostringstream line_kright;
	line_kright
		<< Locale::get( Locale::CONTROLS_STR_KRIGHT )
		<< std::setfill(' ') << std::setw(NAMES_FIELD) << std::left
		<< getname( plcntrl.kright );
	// keyboard left
	std::ostringstream line_kleft;
	line_kleft
		<< Locale::get( Locale::CONTROLS_STR_KLEFT )
		<< std::setfill(' ') << std::setw(NAMES_FIELD) << std::left
		<< getname( plcntrl.kleft );
	// keyboard fire up
	std::ostringstream line_kfireup;
	line_kfireup
		<< Locale::get( Locale::CONTROLS_STR_KFIREUP )
		<< std::setfill(' ') << std::setw(NAMES_FIELD) << std::left
		<< getname( plcntrl.kfireup );
	// keyboard fire down
	std::ostringstream line_kfiredown;
	line_kfiredown
		<< Locale::get( Locale::CONTROLS_STR_KFIREDOWN )
		<< std::setfill(' ') << std::setw(NAMES_FIELD) << std::left
		<< getname( plcntrl.kfiredown );
	// keyboard fire right
	std::ostringstream line_kfireright;
	line_kfireright
		<< Locale::get( Locale::CONTROLS_STR_KFIRERIGHT )
		<< std::setfill(' ') << std::setw(NAMES_FIELD) << std::left
		<< getname( plcntrl.kfireright );
	// keyboard fire left
	std::ostringstream line_kfireleft;
	line_kfireleft
		<< Locale::get( Locale::CONTROLS_STR_KFIRELEFT )
		<< std::setfill(' ') << std::setw(NAMES_FIELD) << std::left
		<< getname( plcntrl.kfireleft );

	// joystick escape
	std::ostringstream line_jescape;
	line_jescape
		<< Locale::get( Locale::CONTROLS_STR_JESCAPE )
		<< std::setfill(' ') << std::setw(NAMES_FIELD) << std::left
		<< getname( false, plcntrl.jescape );
	// joystick bomb
	std::ostringstream line_jbomb;
	line_jbomb
		<< Locale::get( Locale::CONTROLS_STR_JBOMB )
		<< std::setfill(' ') << std::setw(NAMES_FIELD) << std::left
		<< getname( false, plcntrl.jbomb );
	// joystick up/down
	std::ostringstream line_jupdown;
	line_jupdown
		<< Locale::get( Locale::CONTROLS_STR_JUPDOWN )
		<< std::setfill(' ') << std::setw(NAMES_FIELD) << std::left
		<< getname( true, plcntrl.jupdown );
	// joystick right/left
	std::ostringstream line_jrightleft;
	line_jrightleft
		<< Locale::get( Locale::CONTROLS_STR_JRIGHTLEFT )
		<< std::setfill(' ') << std::setw(NAMES_FIELD) << std::left
		<< getname( true, plcntrl.jrightleft );
	// joystick fire up/down
	std::ostringstream line_jfireupdown;
	line_jfireupdown
		<< Locale::get( Locale::CONTROLS_STR_JFIREUPDOWN )
		<< std::setfill(' ') << std::setw(NAMES_FIELD) << std::left
		<< getname( true, plcntrl.jfireupdown );
	// joystick fire right/left
	std::ostringstream line_jfirerightleft;
	line_jfirerightleft
		<< Locale::get( Locale::CONTROLS_STR_JFIRERIGHTLEFT )
		<< std::setfill(' ') << std::setw(NAMES_FIELD) << std::left
		<< getname( true, plcntrl.jfirerightleft );

	// saving the results
	text[CONTROLS_LINE_PLAYER]         = line_player.str();
	text[CONTROLS_LINE_KEYBOARD]       = Locale::get( Locale::CONTROLS_STR_KEYBOARD );
	text[CONTROLS_LINE_KESCAPE]        = line_kescape.str();
	text[CONTROLS_LINE_KBOMB]          = line_kbomb.str();
	text[CONTROLS_LINE_KUP]            = line_kup.str();
	text[CONTROLS_LINE_KDOWN]          = line_kdown.str();
	text[CONTROLS_LINE_KRIGHT]         = line_kright.str();
	text[CONTROLS_LINE_KLEFT]          = line_kleft.str();
	text[CONTROLS_LINE_KFIREUP]        = line_kfireup.str();
	text[CONTROLS_LINE_KFIREDOWN]      = line_kfiredown.str();
	text[CONTROLS_LINE_KFIRERIGHT]     = line_kfireright.str();
	text[CONTROLS_LINE_KFIRELEFT]      = line_kfireleft.str();
	text[CONTROLS_LINE_JOYSTICK]       = Locale::get( Locale::CONTROLS_STR_JOYSTICK );
	text[CONTROLS_LINE_JESCAPE]        = line_jescape.str();
	text[CONTROLS_LINE_JBOMB]          = line_jbomb.str();
	text[CONTROLS_LINE_JUPDOWN]        = line_jupdown.str();
	text[CONTROLS_LINE_JRIGHTLEFT]     = line_jrightleft.str();
	text[CONTROLS_LINE_JFIREUPDOWN]    = line_jfireupdown.str();
	text[CONTROLS_LINE_JFIRERIGHTLEFT] = line_jfirerightleft.str();
	text[CONTROLS_LINE_OK]             = Locale::get( Locale::STR_OK );

	Menu::update();
}
// save controls
void gui::Menu_Controls::save(){
	Config::set_controls( playercntrls );
	Input::reinit();
}
// tests if present config is acceptable, that is: has the needed to let the player exit the game
bool gui::Menu_Controls::acceptable(){
	bool escape = false;
	bool bomb   = false;
	bool updown = false;

	// searching in all available configurations
	foreach( const Config::playercontrols& p, playercntrls ){
		escape = escape
		         || (p.kescape != SDLK_UNKNOWN)
		         || (p.jescape != std::pair<int,int>(-1, -1));
		bomb   = bomb
		         || (p.kbomb != SDLK_UNKNOWN)
		         || (p.jbomb != std::pair<int,int>(-1, -1));
		updown = updown
		         || (p.kup != SDLK_UNKNOWN)
		         || (p.kdown != SDLK_UNKNOWN)
		         || (p.jupdown != std::pair<int,int>(-1, -1));
		if ( escape && bomb && updown )
			break;
	}

	return escape && bomb && updown;
}
// sets plcntrl with pertinent info
void gui::Menu_Controls::resetplcntrl(){
	// if it is the next player, controls are unknown
	if ( playerid == static_cast<int>(playercntrls.size()) ){
		// keyboard info
		plcntrl.kescape              = SDLK_UNKNOWN;
		plcntrl.kimm                 = SDLK_UNKNOWN;
		plcntrl.kbomb                = SDLK_UNKNOWN;
		plcntrl.kup                  = SDLK_UNKNOWN;
		plcntrl.kdown                = SDLK_UNKNOWN;
		plcntrl.kright               = SDLK_UNKNOWN;
		plcntrl.kleft                = SDLK_UNKNOWN;
		plcntrl.kfireup              = SDLK_UNKNOWN;
		plcntrl.kfiredown            = SDLK_UNKNOWN;
		plcntrl.kfireright           = SDLK_UNKNOWN;
		plcntrl.kfireleft            = SDLK_UNKNOWN;
		// joystick info
		plcntrl.jescape              = std::pair<int,int>( -1, -1 );
		plcntrl.jimm                 = std::pair<int,int>( -1, -1 );
		plcntrl.jbomb                = std::pair<int,int>( -1, -1 );
		plcntrl.jupdown              = std::pair<int,int>( -1, -1 );
		plcntrl.jinvertupdown        = false;
		plcntrl.jrightleft           = std::pair<int,int>( -1, -1 );
		plcntrl.jinvertrightleft     = false;
		plcntrl.jfireupdown          = std::pair<int,int>( -1, -1 );
		plcntrl.jinvertfireupdown    = false;
		plcntrl.jfirerightleft       = std::pair<int,int>( -1, -1 );
		plcntrl.jinvertfirerightleft = false;
	}else // else load the present ones
		plcntrl = playercntrls[playerid];
}
// user event: exit
void gui::Menu_Controls::exit(){
	end();
}
// user event: movement
void gui::Menu_Controls::move( int updown ){
	// changing line
	line += updown;
	switch(line){
		case CONTROLS_LINE_SKIP0:
		case CONTROLS_LINE_KEYBOARD:
		case CONTROLS_LINE_SKIP1:
		case CONTROLS_LINE_JOYSTICK:
		case CONTROLS_LINE_SKIP2:
			line += updown;
			break;
	}
	// there is a couple of two consecutive skipped lines, recheck line
	switch(line){
		case CONTROLS_LINE_SKIP0:
		case CONTROLS_LINE_KEYBOARD:
		case CONTROLS_LINE_SKIP1:
		case CONTROLS_LINE_JOYSTICK:
			line += updown;
			break;
	}
	// wrapping
	if ( line < CONTROLS_LINE_MIN )
		line = CONTROLS_LINE_MAX;
	else if ( line > CONTROLS_LINE_MAX )
		line = CONTROLS_LINE_MIN;
}
// user event: change
void gui::Menu_Controls::change( int moreless ){
	switch(line){
		// only player can be changed
		case CONTROLS_LINE_PLAYER:
			playerid = std::min(
				std::min( MAX_PLAYER_AMOUNT-1, static_cast<int>(playercntrls.size()) ),
				std::max( 0, playerid+moreless ) );
			resetplcntrl();
			update();
			break;
	}
}
// user event: accept
void gui::Menu_Controls::accept(){
	switch(line){
		// starting keyboard capture
		case CONTROLS_LINE_KESCAPE:
		case CONTROLS_LINE_KBOMB:
		case CONTROLS_LINE_KUP:
		case CONTROLS_LINE_KDOWN:
		case CONTROLS_LINE_KRIGHT:
		case CONTROLS_LINE_KLEFT:
		case CONTROLS_LINE_KFIREUP:
		case CONTROLS_LINE_KFIREDOWN:
		case CONTROLS_LINE_KFIRERIGHT:
		case CONTROLS_LINE_KFIRELEFT:
			Menu::capture( Menu::CAPTURE_KEY, Menu::CONTROLS_PANE );
			break;
		// starting joystick button capture
		case CONTROLS_LINE_JESCAPE:
		case CONTROLS_LINE_JBOMB:
			Menu::capture( Menu::CAPTURE_JOYSTICK_BUTTON, Menu::CONTROLS_PANE );
			break;
		// starting joystick axis capture
		case CONTROLS_LINE_JUPDOWN:
		case CONTROLS_LINE_JRIGHTLEFT:
		case CONTROLS_LINE_JFIREUPDOWN:
		case CONTROLS_LINE_JFIRERIGHTLEFT:
			Menu::capture( Menu::CAPTURE_JOYSTICK_AXIS, Menu::CONTROLS_PANE );
			break;
		// ending
		case CONTROLS_LINE_OK:
			if ( acceptable() ){
				save();
				end();
			}
			break;
	}
}
// capture ended: saving result
void gui::Menu_Controls::notify( bool success, SDLKey key, int joy, int axis_but, bool invert ){
	// variables to store present configured value that will be freed once reconfigured
	SDLKey prev_key = SDLK_UNKNOWN;
	std::pair<int,int> prev_joybut(-1, -1);
	std::pair<int,int> prev_joyaxis(-1, -1);

	if ( success ){
		// updating
		switch(line){
			// updating keys
			case CONTROLS_LINE_KESCAPE:
				prev_key = plcntrl.kescape;
				plcntrl.kescape = key;
				break;
			case CONTROLS_LINE_KBOMB:
				prev_key = plcntrl.kbomb;
				plcntrl.kbomb = key;
				break;
			case CONTROLS_LINE_KUP:
				prev_key = plcntrl.kup;
				plcntrl.kup = key;
				break;
			case CONTROLS_LINE_KDOWN:
				prev_key = plcntrl.kdown;
				plcntrl.kdown = key;
				break;
			case CONTROLS_LINE_KRIGHT:
				prev_key = plcntrl.kright;
				plcntrl.kright = key;
				break;
			case CONTROLS_LINE_KLEFT:
				prev_key = plcntrl.kleft;
				plcntrl.kleft = key;
				break;
			case CONTROLS_LINE_KFIREUP:
				prev_key = plcntrl.kfireup;
				plcntrl.kfireup = key;
				break;
			case CONTROLS_LINE_KFIREDOWN:
				prev_key = plcntrl.kfiredown;
				plcntrl.kfiredown = key;
				break;
			case CONTROLS_LINE_KFIRERIGHT:
				prev_key = plcntrl.kfireright;
				plcntrl.kfireright = key;
				break;
			case CONTROLS_LINE_KFIRELEFT:
				prev_key = plcntrl.kfireleft;
				plcntrl.kfireleft = key;
				break;
			// updating joystick buttons
			case CONTROLS_LINE_JESCAPE:
				prev_joybut = plcntrl.jescape;
				plcntrl.jescape = std::pair<int,int>(joy, axis_but);
				break;
			case CONTROLS_LINE_JBOMB:
				prev_joybut = plcntrl.jbomb;
				plcntrl.jbomb = std::pair<int,int>(joy, axis_but);
				break;
			// updating joystick axis
			case CONTROLS_LINE_JUPDOWN:
				prev_joyaxis = plcntrl.jupdown;
				plcntrl.jupdown = std::pair<int,int>(joy, axis_but);
				plcntrl.jinvertupdown = invert;
				break;
			case CONTROLS_LINE_JRIGHTLEFT:
				prev_joyaxis = plcntrl.jrightleft;
				plcntrl.jrightleft = std::pair<int,int>(joy, axis_but);
				plcntrl.jinvertrightleft = invert;
				break;
			case CONTROLS_LINE_JFIREUPDOWN:
				prev_joyaxis = plcntrl.jfireupdown;
				plcntrl.jfireupdown = std::pair<int,int>(joy, axis_but);
				plcntrl.jinvertfireupdown = invert;
				break;
			case CONTROLS_LINE_JFIRERIGHTLEFT:
				prev_joyaxis = plcntrl.jfirerightleft;
				plcntrl.jfirerightleft = std::pair<int,int>(joy, axis_but);
				plcntrl.jinvertfirerightleft = invert;
				break;
		}

		// saving
		if ( playerid == static_cast<int>(playercntrls.size()) )
			playercntrls.push_back( plcntrl );
		else
			playercntrls[playerid] = plcntrl;

		// updating mappings
		switch(line){
			// updating key mappings
			case CONTROLS_LINE_KESCAPE:
			case CONTROLS_LINE_KBOMB:
			case CONTROLS_LINE_KUP:
			case CONTROLS_LINE_KDOWN:
			case CONTROLS_LINE_KRIGHT:
			case CONTROLS_LINE_KLEFT:
			case CONTROLS_LINE_KFIREUP:
			case CONTROLS_LINE_KFIREDOWN:
			case CONTROLS_LINE_KFIRERIGHT:
			case CONTROLS_LINE_KFIRELEFT:
				keymappings.erase( prev_key );
				addkeymapping( key, std::pair<int,int>(playerid, line) );
				break;
			// updating joystick button mappings
			case CONTROLS_LINE_JESCAPE:
			case CONTROLS_LINE_JBOMB:
				buttonmappings.erase( prev_joybut );
				addbuttonmapping( std::pair<int,int>(joy, axis_but), std::pair<int,int>(playerid, line) );
				break;
			// updating joystick axis mappings
			case CONTROLS_LINE_JUPDOWN:
			case CONTROLS_LINE_JRIGHTLEFT:
			case CONTROLS_LINE_JFIREUPDOWN:
			case CONTROLS_LINE_JFIRERIGHTLEFT:
				axismappings.erase( prev_joyaxis );
				addaxismapping( std::pair<int,int>(joy, axis_but), std::pair<int,int>(playerid, line) );
				break;
		}

		// updating screen
		resetplcntrl();
		update();
	}
}
