// nvsts is a multi-directional shooter game.
// Copyright (C) 2008 Carles Creus Lopez
//
// nvsts is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>
// or write to the Free Software Foundation, Inc., 51 Franklin Street,
// Fifth Floor, Boston, MA 02110-1301 USA.
//
// You may contact the author at <crazycraft@gmail.com>

#ifndef VECTOR_HPP
#define VECTOR_HPP

#include <cmath>
#include <istream>
#include <ostream>
#include "Real.hpp"

namespace util {

class Vector{
	public:
		// ctor
		Vector( const real x = 0.0, const real y = 0.0 ) :
		 x(x),
		 y(y)
		{
		}
		Vector( const Vector& v ) :
		 x(v.x),
		 y(v.y)
		{
		}

		// get this vector's components
		real& cx(){
			return x;
		}
		real& cy(){
			return y;
		}
		const real& cx() const{
			return x;
		}
		const real& cy() const{
			return y;
		}
		// get this vector's magnitude
		real magnitude() const{
			return std::sqrt( magnitude2() );
		}
		// get this vector's squared magnitude
		real magnitude2() const{
			return x*x + y*y;
		}
		// get this vector's angle
		real angle() const{
			return std::atan2( y, x );
		}

		// set this vector's magnitude
		Vector& magnitude( const real mag ){
			real ang = angle();
			x = mag*std::cos(ang);
			y = mag*std::sin(ang);
			return *this;
		}
		// set this vector's angle
		Vector& angle( const real ang ){
			real mag = magnitude();
			x = mag*std::cos(ang);
			y = mag*std::sin(ang);
			return *this;
		}

		// operators between Vectors
		Vector& operator=( const Vector& v ){
			x = v.x;
			y = v.y;
			return *this;
		}
		Vector& operator+=( const Vector& v ){
			x += v.x;
			y += v.y;
			return *this;
		}
		Vector operator+( const Vector& v ) const{
			Vector res(*this);
			res += v;
			return res;
		}
		Vector& operator-=( const Vector& v ){
			x -= v.x;
			y -= v.y;
			return *this;
		}
		Vector operator-( const Vector& v ) const{
			Vector res(*this);
			res -= v;
			return res;
		}
		Vector operator-() const{
			return Vector( -x, -y );
		}
		bool operator==( const Vector& v ) const{
			return (x == v.x) && (y == v.y);
		}
		bool operator!=( const Vector& v ) const{
			return !(*this == v);
		}
		real operator*( const Vector& v ) const{
			return x*v.x + y*v.y;
		}
		// operators with scalars
		Vector& operator*=( const real k ){
			x *= k;
			y *= k;
			return *this;
		}
		Vector operator*( const real k ) const{
			Vector res(*this);
			res *= k;
			return res;
		}
		Vector& operator/=( const real k ){
			x /= k;
			y /= k;
			return *this;
		}
		Vector operator/( const real k ) const{
			Vector res(*this);
			res /= k;
			return res;
		}

	private:
		// data members
		real x;
		real y;
};

}

// formatted io
extern std::ostream& operator<<( std::ostream& out, const util::Vector& v );
extern std::istream& operator>>( std::istream& in, util::Vector& v );

#endif // Vector.hpp
