// nvsts is a multi-directional shooter game.
// Copyright (C) 2008 Carles Creus Lopez
//
// nvsts is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>
// or write to the Free Software Foundation, Inc., 51 Franklin Street,
// Fifth Floor, Boston, MA 02110-1301 USA.
//
// You may contact the author at <crazycraft@gmail.com>

#ifndef DRAW_HPP
#define DRAW_HPP

#include "../util/Real.hpp"
#include "../util/Vector.hpp"

namespace vis {

// magic id numbers for recording drawings
enum magicID {
	BULLET_ID = 1,
	EXAMINIS_APICULA_ID,
	EXAMINIS_QUEEN_ID,
	LAGVARD_ID,
	PLAYER_ID,
	PLAYER_ID0,
	PROTON_ID,
	REDHOLE_ID,
	SCARED_ID,
	SHIELDED_ID,
	SPAWNED_ID,
	SPAWNSHIP_ID
};

class Draw{
	public:
		// record operations that will be performed
		static void start_record( int id );
		static void end_record();
		static void play_record( int id );

		// drawing line's width
		static void set_thin_lines();
		static void set_normal_lines();

		// free solid-quads drawing context
		static void start_solid_quads();
		static void end_solid_quads();
		// free line-loop drawing context
		static void start_line_loop();
		static void end_line_loop();
		// free line-strip drawing context
		static void start_line_strip();
		static void end_line_strip();
		// free lines drawing context
		static void start_lines();
		static void end_lines();
		// draws a square
		static void square( real side );
		// draws a concave square
		static void concave_square( real side, real inner );
		// draws a rhomb
		static void rhomb( real diagonal );
		// draws a triangle
		static void triangle( real radius );
		// draws a plus
		static void cross_plus( real length );
		// draws a circle (points > 2)
		static void circle( real radius, int points );
		// draws a circular saw (teeth > 2)
		static void circular_saw( real inner_r, real outter_r, int teeth );

		// adds a point to the current path
		static void add_point( const util::Vector& p );
		static void add_point( real x, real y );
		// adds a curve to the current path from previous point to (x3,y3)
		// with (x1,y1) and (x2,y2) as control points, without changing color
		static void add_curve( real x1, real y1, real x2, real y2, real x3, real y3 );
		// adds a curve to the current path from previous point to (x3,y3)
		// with (x1,y1) and (x2,y2) as control points, with a color gradient
		static void add_curve( const util::Vector& p1, real x2, real y2, real x3, real y3 );
		static void add_curve( real x1, real y1, const util::Vector& p2, const util::Vector& p3 );
		static void add_curve( const util::Vector& p1, const util::Vector& p2, const util::Vector& p3 );

		// sets current color
		static void set_color( real r, real g, real b, real a = 1.0 );

		// clears current drawing state matrix
		static void reset();

		// matrix manipulation
		static void translate( const util::Vector& p );
		static void rotate( real angle );
		static void scale( real s );
		static void scale( real sx, real sy );

	private:
		// adds a curve to the current path from previous point to (x3,y3)
		// with (x1,y1) and (x2,y2) as control points, with or without a color gradient
		static void add_curve( const util::Vector& p1, const util::Vector& p2, const util::Vector& p3, bool color );

		// static variables to remember some parameters
		static real previous_color[];
		static real  present_color[];
		static real present_point[];
};

}

#endif // Draw.hpp
