// nvsts is a multi-directional shooter game.
// Copyright (C) 2008 Carles Creus Lopez
//
// nvsts is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>
// or write to the Free Software Foundation, Inc., 51 Franklin Street,
// Fifth Floor, Boston, MA 02110-1301 USA.
//
// You may contact the author at <crazycraft@gmail.com>

#include <cstdlib>
#include <iostream>
#include <algorithm>
#include <SDL_opengl.h>
#include <boost/foreach.hpp>
#define foreach BOOST_FOREACH
#define real_real real,real
#include "../rc/Constants.hpp"
#include "../Main.hpp"
#include "Draw.hpp"
#include "Font.hpp"

#if !defined(FONT_FILE_PREFIX)
# define FONT_FILE_PREFIX FILES_DIRECTORY
#endif // not defined FONT_FILE_PREFIX

#define FONT_NAME PACKAGE_TARNAME ".ttf"

// number of text lines in the screen
#define SCREEN_TEXT_LINES 30

// font face
FTFont *vis::Font::font = NULL;

// initializes the font [depends on: Constants]
void vis::Font::init(){
	font = new FTTextureFont( FONT_FILE_PREFIX FONT_NAME );
	if ( font->Error() ){
		 std::cerr << "FATAL: cannot load the font \"" << FONT_FILE_PREFIX FONT_NAME << "\"" << std::endl;
		 Main::quit( Main::ERROR_FONT );
	}
	font->FaceSize( rc::font_point_size() );
}
// ends the font
void vis::Font::finalize(){
	if ( font != NULL )
		delete font;
}

// ctor, specifying the texttype, the text to render and the color callback
vis::Font::Font( int type, const std::vector<std::string>& text, void (*color) (int) ) :
 frame( false ),
 bottom( 0.0 ),
 top( 0.0 ),
 left( 0.0 ),
 right( 0.0 ),
 color( color ),
 parts(),
 scale( 0.0 ),
 text( text.begin(), text.end() )
{
	// global dimension
	real gwidth = 0.0;
	// all lines will have the same height
	real lineheight = font->Ascender() + -font->Descender();

	// computing real sizes
	foreach( const std::string& s, text ){
		// updating global dimension
		real width = font->Advance( s.c_str() );
		switch(type){
			case GUI_TEXT:
				gwidth += width;
				break;
			case MENU_TEXT:
			case MENU_TEXT_FRAMED:
				gwidth = std::max( gwidth, width );
				break;
		}

		// saving temporal data
		parts.push_back( std::pair<real,real>(width, lineheight) );
	}

	if ( gwidth > 0.0 ){
		// scale factor
		switch(type){
			case GUI_TEXT:
				scale = rc::width() / gwidth;
				break;
			case MENU_TEXT:
			case MENU_TEXT_FRAMED:
				scale = (rc::height()/SCREEN_TEXT_LINES) / (lineheight);
				break;
		}

		// frame
		real margin;
		real aux1;
		real aux2;
		switch(type){
			case MENU_TEXT_FRAMED:
				frame = true;

				margin = lineheight*scale;

				aux1 = lineheight*parts.size()*scale/2.0 + margin;
				bottom = rc::ring_middley() - aux1;
				top    = rc::ring_middley() + aux1;

				aux2 = gwidth*scale/2.0 + margin;
				left  = rc::ring_middlex() - aux2;
				right = rc::ring_middlex() + aux2;
				break;
		}

		// variables for the position
		real positionX0 = 0.0;
		real positionY0 = 0.0;
		real positionX1 = 0.0;
		switch(type){
			case GUI_TEXT:
				positionX1 = rc::ring_x0();
				positionY0 = rc::ring_y1() - lineheight*scale;
				break;
			case MENU_TEXT:
			case MENU_TEXT_FRAMED:
				positionY0 = rc::ring_middley() + lineheight*parts.size()*scale/2.0;
				break;
		}

		// computing positions
		foreach( std::pair<real_real>& p, parts ){
			switch(type){
				case GUI_TEXT:
					positionX0 = positionX1;
					positionX1 = positionX0 + p.first*scale;
					break;
				case MENU_TEXT:
				case MENU_TEXT_FRAMED:
					positionX0 = rc::ring_middlex() - p.first*scale/2.0;
					positionY0 = positionY0 - p.second*scale;
					break;
			}

			p.first  = (positionX0 / scale);
			p.second = (positionY0 / scale) + -font->Descender();
		}
	}
}
// dtor
vis::Font::~Font(){
	parts.clear();
	text.clear();
}
// drawing function
void vis::Font::draw(){
	// drawing frame
	if ( frame ){
		// frame background
		glColor4( rc::menu_background_c_r(), rc::menu_background_c_g(), rc::menu_background_c_b(), rc::menu_background_c_a() );
		glBegin( GL_QUADS );
			glVertex2(  left, bottom );
			glVertex2( right, bottom );
			glVertex2( right,    top );
			glVertex2(  left,    top );
		glEnd();

		// frame margin
		glColor4( rc::menu_margin_c_r(), rc::menu_margin_c_g(), rc::menu_margin_c_b(), rc::menu_margin_c_a() );
		glBegin( GL_LINE_LOOP );
			glVertex2(  left, bottom );
			glVertex2( right, bottom );
			glVertex2( right,    top );
			glVertex2(  left,    top );
		glEnd();
	}

	// the text parts
	int i = 0;
	std::list<std::string>::const_iterator itext = text.begin();
	foreach( const std::pair<real_real>& p, parts ){
		(*color)( i );
		glPushMatrix();
			vis::Draw::scale( scale );
			font->Render( (*itext).c_str(), -1, FTPoint( p.first, p.second ) );
		glPopMatrix();

		++itext;
		++i;
	}
}
