// nvsts is a multi-directional shooter game.
// Copyright (C) 2008 Carles Creus Lopez
//
// nvsts is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>
// or write to the Free Software Foundation, Inc., 51 Franklin Street,
// Fifth Floor, Boston, MA 02110-1301 USA.
//
// You may contact the author at <crazycraft@gmail.com>

#include <ctime>
#include <cstdio>
#include <cstdlib>
#include <string>
#include <iostream>
#include <SDL.h>
#include <SDL_opengl.h>
#include "../rc/Constants.hpp"
#include "../util/Matrix.hpp"
#include "../Config.hpp"
#include "../gui/GUI.hpp"
#include "../gui/Menu.hpp"
#include "../Main.hpp"
#include "Screen.hpp"

// screenshot's number: YYYYMMDDhhmmss
#define SCREENSHOT_NAME         PACKAGE_TARNAME "-screenshot-"
#define SCREENSHOT_NUMBERLENGTH 14
#define SCREENSHOT_EXTENSION    ".bmp"

// available resolutions
// always at least one: the present resolution
std::set< std::pair<int,int> > vis::Screen::available;

// pixel size in game's dimensions
real vis::Screen::pixelsize;

// screen's SDL surface
SDL_Surface *vis::Screen::surf;

// (re)initializes the screen [depends on: Config, GUI]
void vis::Screen::init(){
	// initializing OpenGL
	SDL_GL_SetAttribute( SDL_GL_RED_SIZE, 5 );
	SDL_GL_SetAttribute( SDL_GL_GREEN_SIZE, 5 );
	SDL_GL_SetAttribute( SDL_GL_BLUE_SIZE, 5 );
	SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );

	// starting screen
	surf = SDL_SetVideoMode( Config::get_resx(), Config::get_resy(), Config::get_bitdepth(),
	                         SDL_OPENGL | (Config::get_fullscreen() ? SDL_FULLSCREEN : 0) );
	if ( surf == NULL ){
		std::cerr << "FATAL: cannot start window: " << SDL_GetError() << std::endl;
		Main::quit( Main::ERROR_RESOLUTION );
	}
	// checking double buffered
	int aux;
	if ( SDL_GL_GetAttribute( SDL_GL_DOUBLEBUFFER, &aux ) == -1 )
		aux = 0;
	if ( aux != 1 ){
		std::cerr << "FATAL: cannot start double buffered window" << std::endl;
		Main::quit( Main::ERROR_DBUFFER );
	}

	// getting available screen resolutions
	available.clear();
	available.insert( std::pair<int,int>(Config::get_resx(), Config::get_resy()) );
	SDL_Rect **opt = SDL_ListModes( surf->format, surf->flags );
	if ( opt != NULL ){
		// if we've got a list of resolutions, add them all
		if ( opt != reinterpret_cast<SDL_Rect**>(-1) ){
			for( ; *opt != NULL ; ++opt )
				available.insert( std::pair<int,int>(static_cast<int>((*opt)->w), static_cast<int>((*opt)->h)) );
		}else{
			// else we try for a list of common resolutions
			if ( SDL_VideoModeOK(  480,  360, Config::get_bitdepth(), SDL_OPENGL | SDL_FULLSCREEN ) )
				available.insert( std::pair<int,int>(  480,  360 ) );
			if ( SDL_VideoModeOK(  640,  480, Config::get_bitdepth(), SDL_OPENGL | SDL_FULLSCREEN ) )
				available.insert( std::pair<int,int>(  640,  480 ) );
			if ( SDL_VideoModeOK(  800,  600, Config::get_bitdepth(), SDL_OPENGL | SDL_FULLSCREEN ) )
				available.insert( std::pair<int,int>(  800,  600 ) );
			if ( SDL_VideoModeOK( 1024,  768, Config::get_bitdepth(), SDL_OPENGL | SDL_FULLSCREEN ) )
				available.insert( std::pair<int,int>( 1024,  768 ) );
			if ( SDL_VideoModeOK( 1280, 1024, Config::get_bitdepth(), SDL_OPENGL | SDL_FULLSCREEN ) )
				available.insert( std::pair<int,int>( 1280, 1024 ) );
			if ( SDL_VideoModeOK( 1600, 1200, Config::get_bitdepth(), SDL_OPENGL | SDL_FULLSCREEN ) )
				available.insert( std::pair<int,int>( 1600, 1200 ) );
		}
	}

	// computing present pixel size in game's dimensions
	pixelsize = Config::get_resx()/rc::width();

	// taking out the cursor
	SDL_ShowCursor( SDL_DISABLE );

	// setting window name
	SDL_WM_SetCaption( PACKAGE_STRING, PACKAGE_NAME );

	// starting OpenGL
	glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );
	glClearColor( rc::clear_c_r(), rc::clear_c_g(), rc::clear_c_b(), rc::clear_c_a() );
	if ( Config::get_linesmooth() )
		glEnable( GL_LINE_SMOOTH );
	else
		glDisable( GL_LINE_SMOOTH );
	glEnable( GL_BLEND );
	glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );

	glViewport( 0, 0, Config::get_resx(), Config::get_resy() );
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	glOrtho( 0, static_cast<int>(rc::width()), 0, static_cast<int>(rc::height()), 0, 1.0 );
	glMatrixMode( GL_MODELVIEW );

	// init specific OpenGL data
	rc::initalldrawings();
	gui::GUI::initdrawing();
	gui::Menu::initdrawing();
}
// prepares the screen for drawing
void vis::Screen::start_drawing(){
	glClear( GL_COLOR_BUFFER_BIT );
}
// all drawing has been done
void vis::Screen::end_drawing(){
	SDL_GL_SwapBuffers();
}
// saves a screenshot
void vis::Screen::takescreenshot(){
	// getting time
	time_t now;
	if ( std::time( &now ) == static_cast<time_t>(-1) )
		return;

	struct tm *nows = std::localtime( &now );
	if ( nows == NULL )
		return;

	// printing time
	char number[SCREENSHOT_NUMBERLENGTH+1];
	std::snprintf( number, SCREENSHOT_NUMBERLENGTH+1,
	               "%04d%02d%02d%02d%02d%02d",
	               nows->tm_year + 1900,
	               nows->tm_mon + 1,
	               nows->tm_mday,
	               nows->tm_hour,
	               nows->tm_min,
	               nows->tm_sec );

	// making up the file's name
	std::string filename;
#if defined(WIN32) || defined(_WIN32) || defined(__WIN32__)
	filename = SCREENSHOT_NAME;
	filename += number;
	filename += SCREENSHOT_EXTENSION;
#else // Windows
	char *home = std::getenv( "HOME" );
	if ( home == NULL )
		return;
	filename = home;
	filename += "/" SCREENSHOT_NAME;
	filename += number;
	filename += SCREENSHOT_EXTENSION;
#endif // Not Windows

	util::Matrix<GLubyte> pixels( surf->h, 3*surf->w );

	// reading data
	glReadPixels( 0, 0, surf->w, surf->h, GL_RGB, GL_UNSIGNED_BYTE, &pixels[0][0] );
	if ( glGetError() == GL_NO_ERROR ){
		// flipping lines
		int i = 0;
		int j = pixels.get_rows() - 1;
		while( i < j ){
			for( int k=0 ; k<pixels.get_cols() ; ++k ){
				GLubyte aux = pixels[i][k];
				pixels[i][k] = pixels[j][k];
				pixels[j][k] = aux;
			}
			i++;
			j--;
		}

#if SDL_BYTEORDER == SDL_BIG_ENDIAN
		SDL_Surface *screen = SDL_CreateRGBSurfaceFrom( &pixels[0][0], surf->w, surf->h, 24, pixels.get_cols(),
		                                                0x00ff0000, 0x0000ff00, 0x000000ff, 0x00000000 );
#else // big endian
		SDL_Surface *screen = SDL_CreateRGBSurfaceFrom( &pixels[0][0], surf->w, surf->h, 24, pixels.get_cols(),
		                                                0x000000ff, 0x0000ff00, 0x00ff0000, 0x00000000 );
#endif // little endian
		if ( screen != NULL ){
			SDL_SaveBMP( screen, filename.c_str() );

			SDL_FreeSurface( screen );
		}
	}
}
// gets present pixelsize, i.e. pixels per rc::width()
real vis::Screen::get_pixelsize(){
	return pixelsize;
}
// gets available resolutions
// always at least one: the present resolution
const std::set< std::pair<int,int> >& vis::Screen::get_availableres(){
	return available;
}
